import re
import functools
import collections

import pysam
import numpy as np

import handygenome.tools as tools
import handygenome.workflow as workflow
from handygenome.annotation.annotitem import AnnotItemInfoSingle, AnnotItemFormatSingle, AnnotItemFormatSampledict
import handygenome.variant.ponbams as libponbams
import handygenome.deco as deco


PON_LOGGER_INFO = workflow.get_debugging_logger(title='PONfilter', verbose=False)
PON_LOGGER_DEBUG = workflow.get_debugging_logger(title='PONfilter', verbose=True)
SUFFICIENT_PON_SAMPLE_NUM = 10


# preset filter combinations
def get_preset_filter_germline(
    with_pon=True, 
    pon_samples=None, 
    pon_cohorts=('BGI', 'PCAWG', 'SNULUNG'),
    refver='hg19',

    cutoff_diffbq=-5,
    cutoff_absbq=20,
    cutoff_diffmq=-10,
    cutoff_absmq=40,
    cutoff_clipovlp=2.0,
    cutoff_cliplen=20,
    cutoff_altcount=2,
    cutoff_otherratio=1.5,
    #cutoffs_totaldepth=(10, 100),
    cutoffs_totaldepth=None,
    cutoff_unifpval=0.01,
    cutoff_recurMM=10,
    cutoffs_vaf=None,
):
    # sanity check
    if with_pon:
        if (pon_samples is not None) + (pon_cohorts is not None) != 1:
            raise Exception(
                f'If "with_pon" is True, '
                f'exactly 1 of "pon_samples" and "pon_cohorts" must be set.'
            )

    # main
    result = FilterList()
    if cutoff_diffbq is not None: 
        result.append(DiffMeanBQFilter(cutoff=cutoff_diffbq))
    if cutoff_absbq is not None: 
        result.append(AbsMeanBQFilter(cutoff=cutoff_absbq))
    if cutoff_diffmq is not None: 
        result.append(DiffMeanMQFilter(cutoff=cutoff_diffmq))
    if cutoff_absmq is not None: 
        result.append(AbsMeanMQFilter(cutoff=cutoff_absmq))
    if cutoff_clipovlp is not None: 
        result.append(ClipoverlapFilter(cutoff=cutoff_clipovlp))
    if cutoff_cliplen is not None: 
        result.append(AbsCliplenFilter(cutoff=cutoff_cliplen))
    if cutoff_altcount is not None: 
        result.append(ReadcountFilter(cutoff=cutoff_altcount))
    if cutoff_otherratio is not None: 
        result.append(OthercountRatioFilter(cutoff=cutoff_otherratio))
    if cutoffs_totaldepth is not None: 
        result.append(TotaldepthFilter(cutoff=cutoffs_totaldepth))
    if cutoff_unifpval is not None: 
        result.append(VarposUniformFilter(cutoff=cutoff_unifpval))
    if cutoff_recurMM is not None: 
        result.append(RecurrentMMFilter(cutoff=cutoff_recurMM))
    if cutoffs_vaf is not None:
        result.append(VAFFilter(cutoff=cutoffs_vaf))

    if with_pon:
        result.append(
            get_ponfilter(
                pon_samples=pon_samples, 
                pon_cohorts=pon_cohorts, 
                refver=refver, 
                mode='wgs',
            )
        )

    return result


def get_preset_filter_somatic(
    with_pon=True, 
    pon_samples=None, 
    pon_cohorts=('BGI', 'PCAWG', 'SNULUNG'),
    refver='hg19',

    cutoff_diffbq=-5,
    cutoff_absbq=20,
    cutoff_diffmq=-15,
    cutoff_absmq=40,
    cutoff_clipovlp=2.0,
    cutoff_cliplen=20,
    cutoff_altcount=2,
    cutoff_otherratio=1.5,
    cutoff_totalcount=10,
    cutoff_unifpval=0.05,
):
    # sanity check
    if with_pon:
        if sum([
            pon_samples is not None,
            pon_cohorts is not None,
        ]) != 1:
            raise Exception(f'If "with_pon" is True, exactly 1 of "pon_samples" and "pon_cohorts" must be set.')

    result = FilterList(
        [
#            PopfreqFilter(
#                popnames=("GnomAD", "1000Genomes", "KOREAN", "Korea1K"), 
#                cutoff=0.01,
#            ),
            DiffMeanBQFilter(cutoff=cutoff_diffbq),
            AbsMeanBQFilter(cutoff=cutoff_absbq),
            DiffMeanMQFilter(cutoff=cutoff_diffmq),
            AbsMeanMQFilter(cutoff=cutoff_absmq),
            ClipoverlapFilter(cutoff=cutoff_clipovlp),
            AbsCliplenFilter(cutoff=cutoff_cliplen),
            ReadcountFilter(cutoff=cutoff_altcount),
            OthercountRatioFilter(cutoff=cutoff_otherratio),
            TotaldepthGTFilter(cutoff=cutoff_totalcount),
            VarposUniformFilter(cutoff=cutoff_unifpval),
        ]
    )
    if with_pon:
        result.append(
            get_ponfilter(
                pon_samples=pon_samples, 
                pon_cohorts=pon_cohorts, 
                refver=refver, 
                mode='wgs',
            )
        )

    return result


def get_preset_filter_panelseq(
    with_pon=True, 
    pon_samples=None, 
    pon_cohorts=None,
    refver='hg19',

    cutoff_diffbq=-5,
    cutoff_absbq=20,
    cutoff_diffmq=-15,
    cutoff_absmq=40,
    cutoff_clipovlp=1,
    cutoff_cliplen=20,
    cutoff_altcount=2,
    cutoff_otherratio=1.5,
    cutoff_totalcount=30,
    cutoff_unifpval=None,

    popfreq_names=("GnomAD", "1000Genomes", "KOREAN", "Korea1K"),

    ponfilter_kwargs=dict(),
):
    # sanity check
    if with_pon:
        if sum([
            pon_samples is not None,
            pon_cohorts is not None,
        ]) != 1:
            raise Exception(f'If "with_pon" is True, exactly 1 of "pon_samples" and "pon_cohorts" must be set.')

    result = FilterList(
        [
            DiffMeanBQFilter(cutoff=cutoff_diffbq),
            AbsMeanBQFilter(cutoff=cutoff_absbq),
            DiffMeanMQFilter(cutoff=cutoff_diffmq),
            AbsMeanMQFilter(cutoff=cutoff_absmq),
            ClipoverlapFilter(cutoff=cutoff_clipovlp),
            AbsCliplenFilter(cutoff=cutoff_cliplen),
            ReadcountFilter(cutoff=cutoff_altcount),
            OthercountRatioFilter(cutoff=cutoff_otherratio, ref_length_cutoff=4),
            TotaldepthGTFilter(cutoff=cutoff_totalcount),
        ]
    )

    if cutoff_unifpval is not None:
        result.append(VarposUniformFilter(cutoff=cutoff_unifpval))

    if popfreq_names is not None:
        result.append(
            PopfreqFilter(popnames=popfreq_names, cutoff=0.01)
        )

    if with_pon:
        result.append(
            get_ponfilter(
                pon_samples=pon_samples, 
                pon_cohorts=pon_cohorts, 
                refver=refver, 
                mode='panel_somatic',
                **ponfilter_kwargs,
            )
        )

    return result


# Filter Result classes

class FilterResultBase:
    @property
    def passed(self):
        return self['passed']

    @classmethod
    def init_blank(cls):
        result = cls()
        result['is_passed'] = None
        result['data'] = list()
        return result


class FilterResultInfo(FilterResultBase, AnnotItemInfoSingle):
    meta = {
        'ID': 'filter_result_info', 'Number': 1, 'Type': 'String',
        'Description': 'Result of variant filtering process. Only one item for a variant, not considering each sample.',
    }

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)


class FilterResultFormatUnit(FilterResultBase, AnnotItemFormatSingle):
    pass


class FilterResultFormat(AnnotItemFormatSampledict):
    meta = {
        'ID': 'filter_result_format', 'Number': 1, 'Type': 'String',
        'Description': 'Result of variant filtering process. Separate result for each sample.',
    }
    unit_class = FilterResultFormatUnit

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)


###################################################
# base classes for Filter Objects and FilterList

class FilterBase:
    """"Return value of "check" method: 
        True means the variant is valid
    """
    def __repr__(self):
        return f"<{self.__class__.__name__} ({self.params})>"

    @staticmethod
    def check_value_between(value, lower, upper):
        lower_result = (True if (lower is None) else (value >= lower))
        upper_result = (True if (upper is None) else (value <= upper))
        return lower_result and upper_result


class SamplewiseFilter(FilterBase):
    """"check" method:
        - call signature: def check(self, vp, sampleid=None, allele_index=1)
    """
    @classmethod
    def add_filter(cls, vp, sampleid):
        vp.add_sample_filter(sampleid, cls.__name__)

    def apply(self, vp, sampleid):
        mask = self.check(vp, sampleid)
        if not mask:
            self.add_filter(vp, sampleid)

    def check_show(self, vp, sampleid, allele_index=1):
        mask = self.check(vp, sampleid, allele_index)
        print(mask, self, sep=(" " * 4))

    def _check_meanvalue_difference(
        self, readstats_key, vp, sampleid, allele_index, cutoff, gt,
    ):
        readstats = vp.readstats_dict[sampleid]
        target_value = readstats[readstats_key][allele_index]

        #other_allele_indexes = vp.get_other_allele_indexes(allele_index)
        other_allele_indexes = tuple(sorted(
            set(range(len(vp.vr.alleles))).difference({allele_index})
        ))
        other_value = readstats.get_allele_indexes_average(
            readstats_key, other_allele_indexes
        )

        if np.isnan([target_value, other_value]).any():
            return True
        else:
            diff = target_value - other_value
            if gt:
                return diff >= cutoff
            else:
                return diff < cutoff

    def _check_absvalue(
        self, readstats_key, vp, sampleid, allele_index, cutoff, gt,
    ):
        readstats = vp.readstats_dict[sampleid]
        target_value = readstats[readstats_key][allele_index]

        if np.isnan(target_value):
            return True
        else:
            if gt:
                return target_value >= cutoff
            else:
                return target_value < cutoff


class NonSamplewiseFilter(FilterBase):
    """"check" method:
        - call signature: def check(self, vp, allele_index=1)
    """
    def check_show(self, vp, allele_index=1):
        mask = self.check(vp, allele_index)
        print(mask, self, sep=(" " * 4))


class FilterList(list):
    def __repr__(self):
        repr_string = [repr(x) for x in self]
        return f"<FilterList object ({repr_string})>"

    def get_check_results(self, vp, sampleid, allele_index):
        check_results = list()
        for fi in self:
            if isinstance(fi, NonSamplewiseFilter):
                check_results.append(
                    fi.check(vp, allele_index=allele_index)
                )
            elif isinstance(fi, SamplewiseFilter):
                check_results.append(
                    fi.check(vp, sampleid=sampleid, allele_index=allele_index)
                )
        return check_results

    def check(self, vp, sampleid, allele_index=1, write_result=False):
        check_result = all(self.get_check_results(vp=vp, sampleid=sampleid, allele_index=allele_index))
        if write_result:
            vp.filter['is_passed'] = check_result
        return check_result

    def check_show(self, vp, sampleid, allele_index=1):
        for fi in self:
            if isinstance(fi, NonSamplewiseFilter):
                fi.check_show(vp=vp, allele_index=allele_index)
            elif isinstance(fi, SamplewiseFilter):
                fi.check_show(vp=vp, sampleid=sampleid, allele_index=allele_index)

    def check_any(self, vp, sampleid, allele_index=1):
        return any(self.get_check_results(vp=vp, sampleid=sampleid, allele_index=allele_index))


######################################


class PopfreqFilter(NonSamplewiseFilter):
    def __init__(
        self, 
        popnames=("GnomAD", "1000Genomes", "KOREAN", "Korea1K"), 
        cutoff=0.01,
    ):
        self.params = {
            "popnames": popnames,
            "cutoff": cutoff,
        }

    def check(self, vp, allele_index=1):
        assert allele_index > 0
        tmp = list()
        for popname in self.params["popnames"]:
            freq = vp.popfreq[allele_index - 1].get_freq(popname)
            checkresult = (
                np.isnan(freq)
                or (freq <= self.params["cutoff"])
            )
            tmp.append(checkresult)

        return all(tmp)


class DiffMeanBQFilter(SamplewiseFilter):
    def __init__(self, cutoff=-5):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_meanvalue_difference(
            readstats_key="mean_BQs",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class AbsMeanBQFilter(SamplewiseFilter):
    def __init__(self, cutoff=20):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_absvalue(
            readstats_key="mean_BQs",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class DiffMeanMQFilter(SamplewiseFilter):
    def __init__(self, cutoff=-15):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_meanvalue_difference(
            readstats_key="mean_MQs",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class AbsMeanMQFilter(SamplewiseFilter):
    def __init__(self, cutoff=40):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_absvalue(
            readstats_key="mean_MQs",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class ClipoverlapFilter(SamplewiseFilter):
    def __init__(self, cutoff=1):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        readstats = vp.readstats_dict[sampleid]
        clipoverlap_count = readstats["rppcounts"]["softclip_overlap"]
        target_count = readstats["rppcounts"][allele_index]

        if clipoverlap_count == 0:
            return True
        else:
            ratio = target_count / clipoverlap_count
            return ratio > self.params["cutoff"]


class AvgCliplenFilter(SamplewiseFilter):
    def __init__(self, cutoff=20):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=None):
        readstats = vp.readstats_dict[sampleid]
        avg_cliplen = vp.readstats_dict[sampleid].get_allele_indexes_average(
            key='mean_cliplens', allele_indexes=None,
        )
        return avg_cliplen < self.params['cutoff']


class AbsCliplenFilter(SamplewiseFilter):
    def __init__(self, cutoff=20):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_absvalue(
            readstats_key="mean_cliplens",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=False,
        )


class VarposUniformFilter(SamplewiseFilter):
    def __init__(self, cutoff=0.05):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        readstats = vp.readstats_dict[sampleid]
        pval = readstats["varpos_uniform_pvalues"][allele_index]
        if np.isnan(pval):
            return True
        else:
            return pval >= self.params["cutoff"]


class VarposValueFilter(SamplewiseFilter):
    def __init__(self, cutoff=0.3):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_absvalue(
            readstats_key="mean_varpos_fractions",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class ReadcountFilter(SamplewiseFilter):
    def __init__(self, cutoff=2):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return self._check_absvalue(
            readstats_key="rppcounts",
            vp=vp,
            sampleid=sampleid,
            allele_index=allele_index,
            cutoff=self.params["cutoff"],
            gt=True,
        )


class OthercountRatioFilter(SamplewiseFilter):
    def __init__(self, cutoff=1.5, ref_length_cutoff=None):
        self.params = {
            "cutoff": cutoff,
            'ref_length_cutoff': ref_length_cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        if self.params['ref_length_cutoff'] is not None:
            if len(vp.vcfspec.ref) >= self.params['ref_length_cutoff']:
                return True

        readstats = vp.readstats_dict[sampleid]
        target_count = readstats["rppcounts"][allele_index]
        other_count = readstats["rppcounts"][-1]

        if other_count == 0:
            return True
        else:
            ratio = target_count / other_count
            return ratio > self.params["cutoff"]


class TotaldepthGTFilter(SamplewiseFilter):
    def __init__(self, cutoff=20):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=None, exclude_other=False):
        return vp.readstats_dict[sampleid].get_total_rppcount(
            exclude_other=exclude_other,
        ) >= self.params['cutoff']


class TotaldepthLTFilter(SamplewiseFilter):
    def __init__(self, cutoff=200):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=None, exclude_other=False):
        return vp.readstats_dict[sampleid].get_total_rppcount(
            exclude_other=exclude_other,
        ) <= self.params['cutoff']


class TotaldepthFilter(SamplewiseFilter):
    def __init__(self, cutoff=(20, 200)):
        self.params = {
            "cutoff_lower": cutoff[0],
            "cutoff_upper": cutoff[1],
        }

    def check(self, vp, sampleid, allele_index=None, exclude_other=False):
        depth = vp.readstats_dict[sampleid].get_total_rppcount(exclude_other=exclude_other)
        return self.check_value_between(
            depth, self.params["cutoff_lower"], self.params["cutoff_upper"],
        )


class RecurrentMMFilter(SamplewiseFilter):
    def __init__(self, cutoff=10):
        self.params = {
            "cutoff": cutoff,
        }

    def check(self, vp, sampleid, allele_index=1):
        return (
            vp.readstats_dict[sampleid]['recurrent_mNM'][allele_index] 
            <= self.params['cutoff']
        )


class VAFFilter(SamplewiseFilter):
    def __init__(self, cutoff=(None, None)):
        self.params = {
            "cutoff_lower": cutoff[0],
            "cutoff_upper": cutoff[1],
        }

    def check(self, vp, sampleid, allele_index=1, exclude_other=False):
        vaf = vp.get_vaf(sampleid, allele_index=allele_index, exclude_other=exclude_other)
        return self.check_value_between(
            vaf, self.params["cutoff_lower"], self.params["cutoff_upper"],
        )


############################

# PON filter

@deco.get_deco_arg_choices({'mode': ('wgs', 'panel_germline', 'panel_somatic')})
@deco.get_deco_num_set_differently(('pon_samples', 'pon_cohorts'), 1)
def get_ponfilter(pon_samples=None, pon_cohorts=None, refver=None, mode='wgs', **kwargs):
    if pon_cohorts is not None:
        pon_samples = libponbams.get_pon_sample_names(pon_cohorts, refver)
    else:
        pon_samples = list(pon_samples)

    if mode == 'panel_germline':
        return PonFilterPanelseqGermline(samples=pon_samples, **kwargs)
    elif mode == 'panel_somatic':
        return PonFilterPanelseqSomatic(samples=pon_samples, **kwargs)
    elif mode == 'wgs':
        return PonFilterWGS(samples=pon_samples, **kwargs)


class PonFilterBase(SamplewiseFilter):
    default_cache = {
        'samples_with_low_depth': None,

        'lower_sample_fraction': None,
        'lower_mean': None,
        'lower_stdev': None,
        'lower_cutoff': None,
        'lower_different': None,

        'upper_sample_fraction': None,
        'upper_mean': None,
        'upper_stdev': None,
        'upper_cutoff': None,
        'upper_different': None,

        'is_germline': None,
        'het_vaf_mean': None,
        'germline_fraction_among_upper': None,
        'germline_fraction_among_all': None,

        'is_global': None,

        'upper_result': None,
        
        'total_result': None,
    }

    def __repr__(self):
        infostr = ", ".join(
            [
                f"samples={self.samples}",
                f"params={self.params}",
            ]
        )
        return f"<{self.__class__.__name__} ({infostr})>"

    @staticmethod
    def get_vaf_array(allele_index, readstats_dict, exclude_other):
        """Returns:
        np.ndarray composed of VAFs of PON samples, where 
            "excluded_pon_samples" are excluded and 
            nan values are removed
        Result is sorted in ascending order.
        """
        pon_vafs = list()
        for sampleid, readstats in readstats_dict.items():
#            readstats = readstats_dict[sampleid]
#            denom = readstats.get_total_rppcount()
#            if denom == 0:
#                vaf = np.nan
#            else:
#                numer = readstats["rppcounts"][allele_index]
#                vaf = numer / denom

            vaf = readstats_dict[sampleid].get_vaf(alleleclass=allele_index, exclude_other=exclude_other)
            if not np.isnan(vaf):
                pon_vafs.append(vaf)

        pon_vafs.sort()
        pon_vaf_array = np.array(pon_vafs)

        return pon_vaf_array

    def get_rppcount_array(self, allele_index, readstats_dict):
        return np.array([
            readstats["rppcounts"][allele_index]
            for readstats in readstats_dict.values()
        ])

    def get_used_pon_samples(self, query_sampleid, excluded_pon_samples, exclude_query):
        if excluded_pon_samples is None:
            excluded_pon_samples = set()
        else:
            excluded_pon_samples = set(excluded_pon_samples)

        if exclude_query:
            excluded_pon_samples.add(query_sampleid)

        used_pon_samples = set(self.samples).difference(excluded_pon_samples)
#        if len(used_pon_samples) < self.params["valid_sample_num_cutoff"]:
#            self.logger.warning(
#                f'The number of valid PON samples ({len(used_pon_samples)}) is '
#                f'less than the parameter "valid_sample_num_cutoff" ({self.params["valid_sample_num_cutoff"]}).'
#            )

        return used_pon_samples

    def set_common_params(self, vp, allele_index, query_sampleid, excluded_pon_samples, exclude_query, pon_depth_cutoff, exclude_other):
        # init used_pon_samples
        used_pon_samples = self.get_used_pon_samples(query_sampleid, excluded_pon_samples, exclude_query)
        if not used_pon_samples.issubset(set(vp.readstats_dict.keys())):
            raise Exception(f'Not all of PON sample IDs are included in the VariantPlus object.')

        # additional PON sample removal based on total read depth 
        pon_samples_with_low_depth = set()
        for sampleid in used_pon_samples:
            if vp.readstats_dict[sampleid].get_total_rppcount(exclude_other=exclude_other) < pon_depth_cutoff:
                pon_samples_with_low_depth.add(sampleid)
        used_pon_samples.difference_update(pon_samples_with_low_depth)
        if self.params['save_cache']:
            self.cache['samples_with_low_depth'] = pon_samples_with_low_depth

        # make readstats_dict
        readstats_dict = {
            sampleid: vp.readstats_dict[sampleid] 
            for sampleid in used_pon_samples
        }

        # others
        pon_vaf_array = self.get_vaf_array(allele_index, readstats_dict, exclude_other)  # nan is removed from "pon_vaf_array"
        query_vaf = vp.get_vaf(query_sampleid, allele_index, exclude_other=exclude_other)
        #pon_rppcount_array = self.get_rppcount_array(allele_index, readstats_dict)
        #query_rppcount = vp.readstats_dict[sampleid]["rppcounts"][allele_index]

        return used_pon_samples, readstats_dict, pon_vaf_array, query_vaf

    def reset_cache(self):
        self.cache = self.__class__.default_cache.copy()

    def bisect_vafs(self, pon_vaf_array):
        lower_selector = pon_vaf_array <= self.params['bisect_cutoff']
        return pon_vaf_array[lower_selector], pon_vaf_array[~lower_selector]

    @staticmethod
    def get_deviation(values, mode):
        if mode == 'stdev':
            return np.std(values)
        elif mode == 'mean_mad':
            return tools.mean_mad(values)
        elif mode == 'median_mad':
            return tools.median_mad(values)

    def check_greater_than_lower(
        self, query_vaf, pon_vafs_lower, pon_vaf_array, mode='mean_mad',
    ):
        assert mode in ('stdev', 'mean_mad', 'median_mad')

        sample_fraction = len(pon_vafs_lower) / len(pon_vaf_array)
        if sample_fraction <= self.params['lower_sample_fraction']:
            mean = None
            stdev = None
            cutoff = None
            is_different = True
        else:
            mean = np.mean(pon_vafs_lower)
            stdev = self.get_deviation(pon_vafs_lower, mode)
            cutoff = min(
                0.4,  # 221231
                max(
                    mean + (self.params['lower_stdev_factor'] * stdev),
                    mean * self.params['lower_ratio']
                )
            )
            is_different = query_vaf >= cutoff

        if self.params['save_cache']:
            self.cache['lower_sample_fraction'] = sample_fraction
            self.cache['lower_mean'] = mean
            self.cache['lower_stdev'] = stdev
            self.cache['lower_cutoff'] = cutoff
            self.cache['lower_different'] = is_different

        return mean, stdev, cutoff, is_different

    def check_different_to_upper(
        self, query_vaf, pon_vafs_upper, pon_vaf_array, mode='mean_mad',
    ):
        assert mode in ('stdev', 'mean_mad', 'median_mad')

        sample_fraction = len(pon_vafs_upper) / len(pon_vaf_array)
        if sample_fraction <= self.params['upper_sample_fraction']:
            mean = None
            stdev = None
            cutoff = None
            is_different = True
        else:
            mean = np.mean(pon_vafs_upper)
            stdev = self.get_deviation(pon_vafs_upper, mode)
            cutoff = max(
                mean + (self.params['upper_stdev_factor'] * stdev),
                mean * self.params['upper_ratio']
            )
            is_different = query_vaf >= cutoff

        if self.params['save_cache']:
            self.cache['upper_sample_fraction'] = sample_fraction
            self.cache['upper_mean'] = mean
            self.cache['upper_stdev'] = stdev
            self.cache['upper_cutoff'] = cutoff
            self.cache['upper_different'] = is_different

        return mean, stdev, cutoff, is_different

    def check_global(self, pon_vafs_upper, pon_vaf_array):
        result = (len(pon_vafs_upper) / len(pon_vaf_array)) > self.params['global_sample_ratio_cutoff']
        if self.params['save_cache']:
            self.cache['is_global'] = result

        return result


class PonFilterPanelseqOld(PonFilterBase):
    @deco.get_deco_arg_choices({"mode": ("mean", "max", "median")})
    def __init__(
        self,
        samples,
        verbose=False,

        mode="mean",
        #valid_sample_num_cutoff=20,
        min_germline_vaf=0.2,
        germline_sample_ratio_cutoff=0.1,
        # max_noise_vaf=0.1,
        lowest_subset_fraction=0.7,
        lowest_subset_num_cutoff=10,
        lowest_subset_snr_cutoff=5,
        nearby_ratio=1.3,
        #nearby_subset_fraction_cutoff=0.3,
        nearby_subset_num_cutoff=5,
        nearby_subset_snr_cutoff=3,

        pon_depth_cutoff=30,
    ):
        # set logger
        self.logger = (PON_LOGGER_DEBUG if verbose else PON_LOGGER_INFO)

        # sanity check
        if nearby_ratio < 1:
            raise Exception(
                f'"nearby_ratio" argument must be equal to or greater than 1.'
            )

        # main
        self.samples = samples
        self.params = {
            #"valid_sample_num_cutoff": valid_sample_num_cutoff,
            "min_germline_vaf": min_germline_vaf,
            "germline_sample_ratio_cutoff": germline_sample_ratio_cutoff,
            # "max_noise_vaf": max_noise_vaf,
            "lowest_subset_fraction": lowest_subset_fraction,
            "lowest_subset_num_cutoff": lowest_subset_num_cutoff,
            "lowest_subset_snr_cutoff": lowest_subset_snr_cutoff,
            "nearby_ratio": nearby_ratio,
            #"nearby_subset_fraction_cutoff": nearby_subset_fraction_cutoff,
            "nearby_subset_num_cutoff": nearby_subset_num_cutoff,
            "nearby_subset_snr_cutoff": nearby_subset_snr_cutoff,
            "mode": mode,
            "pon_depth_cutoff": pon_depth_cutoff,
        }

    def check(
        self,
        vp,
        sampleid,
        allele_index=1,
        excluded_pon_samples=None,
        exclude_query=True,
        do_germlinepat=True,
        do_lowest=True,
        do_nearby=True,
        exclude_other=False,
    ):
        # set params
        used_pon_samples, readstats_dict, pon_vaf_array, query_vaf = self.set_common_params(
            vp, allele_index, sampleid, excluded_pon_samples, exclude_query, self.params["pon_depth_cutoff"], exclude_other,
        )
        # main
        if np.isnan(query_vaf):
            return True

        subtests = list()
        if do_germlinepat:
            subtests.append(
                not self.check_germline_pattern(pon_vaf_array)
            )
        if do_lowest:
            subtests.append(
                # is_gt_noise = self.check_greater_than_lowest_count(query_rppcount, pon_rppcount_array)
                self.check_greater_than_lowest_vaf(query_vaf, pon_vaf_array)
            )
        if do_nearby:
            subtests.append(self.check_greater_than_nearby(query_vaf, pon_vaf_array))

        return all(subtests)

    ###################

    def get_lowest_summary_count(self, pon_rppcount_array):
        rppcounts_nonzero = pon_rppcount_array[pon_rppcount_array > 0]
        subset_num = int(len(rppcounts_nonzero) * self.params["subset_fraction"])

        if subset_num >= self.params["subset_num_cutoff"]:
            subset = sorted(rppcounts_nonzero)[:subset_num]
            noise_summary = getattr(np, self.params["mode"])(subset)
        else:
            noise_summary = None

        return noise_summary

    def check_greater_than_lowest_count(self, query_rppcount, pon_rppcount_array):
        noise_summary = self.get_lowest_summary_count(pon_rppcount_array)
        if noise_summary is None:
            return True
        else:
            snr = query_rppcount / noise_summary
            return snr >= self.params["lowest_subset_snr_cutoff"]

    def get_lowest_summary_vaf(self, pon_vaf_array):
        subset_num = int(pon_vaf_array.shape[0] * self.params["lowest_subset_fraction"])
        if subset_num == 0:
            lowest_summary = None
        else:
            subset = pon_vaf_array[:subset_num]
            lowest_summary = getattr(np, self.params["mode"])(subset)

        return lowest_summary

    def check_greater_than_lowest_vaf(self, query_vaf, pon_vaf_array):
        #assert not np.isnan(query_vaf)
        lowest_summary = self.get_lowest_summary_vaf(pon_vaf_array)
        self.logger.debug(f"lowest_summary {lowest_summary}")

        if (lowest_summary is None) or (lowest_summary == 0):
            return True
        else:
            snr = query_vaf / lowest_summary
            self.logger.debug(f"snr {snr}")

            return snr >= self.params["lowest_subset_snr_cutoff"]

    def check_greater_than_nearby(self, query_vaf, pon_vaf_array):
        #assert not np.isnan(query_vaf)
        upper_bound = query_vaf * self.params["nearby_ratio"]
        lower_bound = query_vaf / self.params["nearby_ratio"]
        vafs_nearby_query = pon_vaf_array[
            np.logical_and(pon_vaf_array <= upper_bound, pon_vaf_array >= lower_bound)
        ]
        self.logger.debug(f"upper_bound, {upper_bound}")
        self.logger.debug(f"lower_bound, {lower_bound}")
        self.logger.debug(f"vafs_nearby_query {vafs_nearby_query}")

        if len(vafs_nearby_query) == 0:
            return True
        else:
            #if len(vafs_nearby_query) >= len(pon_vafs) * self.params["nearby_subset_fraction_cutoff"]:
            if len(vafs_nearby_query) >= self.params["nearby_subset_num_cutoff"]:
                nearby_summary = getattr(np, self.params["mode"])(vafs_nearby_query)
                self.logger.debug(f"nearby_summary {nearby_summary}")

                if nearby_summary == 0:
                    return True
                else:
                    snr = query_vaf / nearby_summary
                    self.logger.debug(f'snr {snr}')
                    return snr >= self.params["nearby_subset_snr_cutoff"]
            else:
                return True

    def check_germline_pattern(self, pon_vaf_array):
        if pon_vaf_array.shape[0] == 0:
            self.logger.debug(f"germline_sample_ratio {np.nan}")
            return True
        else:
            n_germline_vaf = (pon_vaf_array >= self.params["min_germline_vaf"]).sum()
            germline_sample_ratio = n_germline_vaf / pon_vaf_array.shape[0]
            self.logger.debug(f"germline_sample_ratio {germline_sample_ratio}")
            return germline_sample_ratio >= self.params["germline_sample_ratio_cutoff"]


class PonFilterWGS(PonFilterBase):
    default_cache = PonFilterBase.default_cache 

    def __init__(
        self,
        samples,
        verbose=False,
        save_cache=False,

        check_global=True,

        bisect_cutoff=0.15,
        lower_stdev_factor=3,
        lower_ratio=3,
        lower_sample_fraction=0.2,
        upper_stdev_factor=3,
        upper_ratio=3,
        upper_sample_fraction=0.2,

        germline_het_range=(0.3, 0.7),
        germline_hom_range=(0.9, 1),
        het_vaf_mean_range=(0.4, 0.6),
        germline_fraction_among_upper_cutoff=0.8,
        germline_fraction_among_all_cutoff=0.1,

        global_sample_ratio_cutoff=0.9,

        pon_depth_cutoff=10,
    ):
        # set logger
        self.logger = (PON_LOGGER_DEBUG if verbose else PON_LOGGER_INFO)

        # sanity check

        # main
        self.samples = samples
        self.params = {
            'save_cache': save_cache,
            'check_global': check_global,

            'bisect_cutoff': bisect_cutoff,
            'lower_stdev_factor': lower_stdev_factor,
            'lower_ratio': lower_ratio,
            'lower_sample_fraction': lower_sample_fraction,
            'upper_stdev_factor': upper_stdev_factor,
            'upper_ratio': upper_ratio,
            'upper_sample_fraction': upper_sample_fraction,

            'germline_het_range': germline_het_range,
            'germline_hom_range': germline_hom_range,
            'het_vaf_mean_range': het_vaf_mean_range,
            'germline_fraction_among_upper_cutoff': germline_fraction_among_upper_cutoff,
            'germline_fraction_among_all_cutoff': germline_fraction_among_all_cutoff,

            'global_sample_ratio_cutoff': global_sample_ratio_cutoff,

            'pon_depth_cutoff': pon_depth_cutoff,
        }

    def check(
        self,
        vp,
        sampleid,
        allele_index=1,
        excluded_pon_samples=None,
        exclude_query=True,
        exclude_other=False,
    ):
        # init cache
        if self.params['save_cache']:
            self.reset_cache()
        # set params
        used_pon_samples, readstats_dict, pon_vaf_array, query_vaf = self.set_common_params(
            vp, allele_index, sampleid, excluded_pon_samples, exclude_query, self.params["pon_depth_cutoff"], exclude_other,
        )
        # main
        if np.isnan(query_vaf):
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        elif len(pon_vaf_array) == 0:
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        else:
            pon_vafs_lower, pon_vafs_upper = self.bisect_vafs(pon_vaf_array)

            lower_mean, lower_stdev, lower_cutoff, lower_different = self.check_greater_than_lower(query_vaf, pon_vafs_lower, pon_vaf_array)
            upper_mean, upper_stdev, upper_cutoff, upper_different = self.check_different_to_upper(query_vaf, pon_vafs_upper, pon_vaf_array)
            is_germline = self.check_is_germline(pon_vafs_upper, pon_vaf_array)
            is_global = self.check_global(pon_vafs_upper, pon_vaf_array)

            if self.params['check_global']:
                upper_result = self.get_upper_result(upper_different, is_germline, is_global)
            else:
                upper_result = self.get_upper_result_without_global(upper_different, is_germline)

            total_result = self.get_total_result(lower_different, upper_result)

            return total_result

    ###########################

    def check_is_germline(self, pon_vafs_upper, pon_vaf_array):
        if len(pon_vafs_upper) == 0:
            het_vaf_mean = None
            germline_fraction_among_upper = None
            germline_fraction_among_all = None
            result = False
        else:
            upper_vaf_is_het = np.logical_and(
                pon_vafs_upper >= self.params['germline_het_range'][0],
                pon_vafs_upper <= self.params['germline_het_range'][1],
            )
            upper_vaf_is_hom = np.logical_and(
                pon_vafs_upper >= self.params['germline_hom_range'][0],
                pon_vafs_upper <= self.params['germline_hom_range'][1],
            )
            upper_vaf_is_germline = np.logical_or(upper_vaf_is_het, upper_vaf_is_hom)

            if not upper_vaf_is_het.any():
                het_vaf_mean = np.nan
            else:
                het_vaf_mean = pon_vafs_upper[upper_vaf_is_het].mean()

            germline_upper_sample_count = upper_vaf_is_germline.sum()
            germline_fraction_among_upper = germline_upper_sample_count / len(pon_vafs_upper)
            germline_fraction_among_all = germline_upper_sample_count / len(pon_vaf_array)

            result = (
                (
                    np.isnan(het_vaf_mean)
                    or (
                        het_vaf_mean >= self.params['het_vaf_mean_range'][0]
                        and het_vaf_mean <= self.params['het_vaf_mean_range'][1]
                    )
                )
                and germline_fraction_among_upper >= self.params['germline_fraction_among_upper_cutoff']
                and germline_fraction_among_all >= self.params['germline_fraction_among_all_cutoff']
            )

        if self.params['save_cache']:
            self.cache['het_vaf_mean'] = het_vaf_mean
            self.cache['germline_fraction_among_upper'] = germline_fraction_among_upper
            self.cache['germline_fraction_among_all'] = germline_fraction_among_all
            self.cache['is_germline'] = result

        return result

    def get_upper_result(self, upper_different, is_germline, is_global):
        if is_global:
            result = False
        elif is_germline:
            result = True
        else:
            result = upper_different

        if self.params['save_cache']:
            self.cache['upper_result'] = result

        return result

    def get_upper_result_without_global(self, upper_different, is_germline):
        if is_germline:
            result = True
        else:
            result = upper_different

        if self.params['save_cache']:
            self.cache['upper_result'] = result

        return result

    def get_total_result(self, lower_different, upper_result):
        result = (lower_different and upper_result)
        if self.params['save_cache']:
            self.cache['total_result'] = result

        return result


class PonFilterPanelseqGermline(PonFilterBase):
    default_cache = PonFilterBase.default_cache 

    def __init__(
        self,
        samples,
        verbose=False,
        save_cache=False,

        check_global=True,

        bisect_cutoff=0.2,
        lower_stdev_factor=1.5,
        lower_ratio=1.5,
        lower_sample_fraction=0.2,
        upper_stdev_factor=1.2,
        upper_ratio=1.2,
        upper_sample_fraction=0.2,

        #germline_het_range=(0.4, 0.6),
        #germline_hom_range=(0.9, 1),
        global_sample_ratio_cutoff=0.9,

        pon_depth_cutoff=50,
    ):
        # set logger
        self.logger = (PON_LOGGER_DEBUG if verbose else PON_LOGGER_INFO)

        # sanity check

        # main
        self.samples = samples
        self.params = {
            'save_cache': save_cache,
            'check_global': check_global,

            'bisect_cutoff': bisect_cutoff,
            'lower_stdev_factor': lower_stdev_factor,
            'lower_ratio': lower_ratio,
            'lower_sample_fraction': lower_sample_fraction,
            'upper_stdev_factor': upper_stdev_factor,
            'upper_ratio': upper_ratio,
            'upper_sample_fraction': upper_sample_fraction,

            #'germline_het_range': germline_het_range,
            #'germline_hom_range': germline_hom_range,
            "global_sample_ratio_cutoff": global_sample_ratio_cutoff,

            "pon_depth_cutoff": pon_depth_cutoff,
        }

    def check(
        self,
        vp,
        sampleid,
        allele_index=1,
        excluded_pon_samples=None,
        exclude_query=True,
        exclude_other=False,
    ):
        # init cache
        if self.params['save_cache']:
            self.reset_cache()
        # set params
        used_pon_samples, readstats_dict, pon_vaf_array, query_vaf = self.set_common_params(
            vp, allele_index, sampleid, excluded_pon_samples, exclude_query, self.params["pon_depth_cutoff"], exclude_other,
        )
        # main
        if np.isnan(query_vaf):
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        elif len(pon_vaf_array) == 0:
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        else:
            pon_vafs_lower, pon_vafs_upper = self.bisect_vafs(pon_vaf_array)

            lower_mean, lower_stdev, lower_cutoff, lower_different = self.check_greater_than_lower(query_vaf, pon_vafs_lower, pon_vaf_array)
            is_global = self.check_global(pon_vafs_upper, pon_vaf_array)

            if self.params['check_global']:
                total_result = self.get_total_result(lower_different, is_global)
            else:
                total_result = self.get_total_result_without_global(lower_different)

            return total_result

    ###########################

    def get_total_result(self, lower_different, is_global):
        result = (lower_different and (not is_global))
        if self.params['save_cache']:
            self.cache['total_result'] = result

        return result

    def get_total_result_without_global(self, lower_different):
        result = lower_different
        if self.params['save_cache']:
            self.cache['total_result'] = result

        return result


class PonFilterPanelseqSomatic(PonFilterBase):
    default_cache = PonFilterBase.default_cache 

    def __init__(
        self,
        samples,
        verbose=False,
        save_cache=False,

        check_global=True,

        bisect_cutoff=0.2,
        lower_stdev_factor=1.5,
        lower_ratio=1.5,
        lower_sample_fraction=0.2,
        upper_stdev_factor=1.2,
        upper_ratio=1.2,
        upper_sample_fraction=0.2,

        #germline_het_range=(0.4, 0.6),
        #germline_hom_range=(0.9, 1),
        germline_sample_ratio_range=(0.2, 1),
        global_sample_ratio_cutoff=0.9,

        pon_depth_cutoff=50,
    ):
        # set logger
        self.logger = (PON_LOGGER_DEBUG if verbose else PON_LOGGER_INFO)

        # sanity check

        # main
        self.samples = samples
        self.params = {
            'save_cache': save_cache,
            'check_global': check_global,

            'bisect_cutoff': bisect_cutoff,
            'lower_stdev_factor': lower_stdev_factor,
            'lower_ratio': lower_ratio,
            'lower_sample_fraction': lower_sample_fraction,
            'upper_stdev_factor': upper_stdev_factor,
            'upper_ratio': upper_ratio,
            'upper_sample_fraction': upper_sample_fraction,

            #'germline_het_range': germline_het_range,
            #'germline_hom_range': germline_hom_range,
            'germline_sample_ratio_range': germline_sample_ratio_range,
            'global_sample_ratio_cutoff': global_sample_ratio_cutoff,

            "pon_depth_cutoff": pon_depth_cutoff,
        }

    def check(
        self,
        vp,
        sampleid,
        allele_index=1,
        excluded_pon_samples=None,
        exclude_query=True,
        exclude_other=False,
    ):
        # init cache
        if self.params['save_cache']:
            self.reset_cache()
        # set params
        used_pon_samples, readstats_dict, pon_vaf_array, query_vaf = self.set_common_params(
            vp, allele_index, sampleid, excluded_pon_samples, exclude_query, self.params["pon_depth_cutoff"], exclude_other,
        )
        # main
        if np.isnan(query_vaf):
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        elif len(pon_vaf_array) == 0:
            if self.params['save_cache']:
                self.cache['total_result'] = True
            return True
        else:
            pon_vafs_lower, pon_vafs_upper = self.bisect_vafs(pon_vaf_array)

            lower_mean, lower_stdev, lower_cutoff, lower_different = self.check_greater_than_lower(query_vaf, pon_vafs_lower, pon_vaf_array)
            is_germline = self.check_is_germline(pon_vafs_upper, pon_vaf_array)
            is_global = self.check_global(pon_vafs_upper, pon_vaf_array)

            if self.params['check_global']:
                total_result = self.get_total_result(lower_different, is_germline, is_global)
            else:
                total_result = self.get_total_result_without_global(lower_different, is_germline)

            return total_result

    ###########################

    def check_is_germline(self, pon_vafs_upper, pon_vaf_array):
        upper_sample_ratio = len(pon_vafs_upper) / len(pon_vaf_array)
        result = (
            upper_sample_ratio >= self.params['germline_sample_ratio_range'][0]
            and upper_sample_ratio <= self.params['germline_sample_ratio_range'][1]
        )
        if self.params['save_cache']:
            self.cache['is_germline'] = result

        return result

    def get_total_result(self, lower_different, is_germline, is_global):
        result = (lower_different and (not is_germline) and (not is_global))
        if self.params['save_cache']:
            self.cache['total_result'] = result

        return result

    def get_total_result_without_global(self, lower_different, is_germline):
        result = (lower_different and (not is_germline))
        if self.params['save_cache']:
            self.cache['total_result'] = result

        return result



############################



#def get_transcript_subtype_filter(key):
#    def vpfilter(vp):
#        return any(
#            feature["transcript_subtype_flags"][key]
#            for feature in vp.annotdb.transcript_canon_ovlp.values()
#        )
#
#    return vpfilter
#
#
#CODING_GENE_INVOLVED = get_transcript_subtype_filter("is_coding")
#
#
#def get_consequence_filter(key):
#    def vpfilter(vp):
#        return any(
#            feature["consequence_flags"][key]
#            for feature in vp.annotdb.transcript_canon_ovlp.values()
#        )
#
#    return vpfilter
#
#
#PROTEIN_ALTERING = get_consequence_filter("is_protein_altering")
#NOT_PROTEIN_ALTERING = get_consequence_filter("is_not_protein_altering")
#
#
#def get_genename_filter(gene_list, canonical_only=True):
#    gene_set = set(gene_list)
#    if canonical_only:
#
#        def vpfilter(vp):
#            vp_genes = set(vp.get_gene_names(canonical_only=True))
#            return not vp_genes.isdisjoint(gene_set)
#
#    else:
#
#        def vpfilter(vp):
#            vp_genes = set(vp.get_gene_names(canonical_only=False))
#            return not vp_genes.isdisjoint(gene_set)
#
#    return vpfilter





import itertools
import importlib

top_package_name = __name__.split('.')[0]
module_getalleleclass_rp = importlib.import_module(top_package_name + '.ria.get_alleleclass_from_readplus')

#from .get_alleleclass_from_readplus import get_seq_REFrange_from_read
#from .get_alleleclass_from_readplus import get_alleleclass_from_read
#from .get_alleleclass_from_readplus import check_read_gives_ambiguous_information
#from .get_alleleclass_from_readplus import generate_pos0_flanking_variant_if_missing
#from .get_alleleclass_from_readplus import get_pos0_flanking_variant


def check_containing_ambiguous_read(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	'''
	Returns True if one or more of the reads give ambiguous information.
	Being ambiguous means partially overlapping REF positions.
	'''
	pos0_before_var, pos0_after_var = \
			module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	for rp in (rpp.rp1, rpp.rp2):
		if module_getalleleclass_rp.check_read_gives_ambiguous_information(
				rp.read, rp.pairs_dict, CHROM, POS0, REF, ALT,
				pos0_before_var = pos0_before_var,
				pos0_after_var = pos0_after_var,
				):
			return True
	return False


def pick_from_two_values(val_list):
	'''
	- If both None, return None
	- If only one is None, return the non-None value
	- If both non-None and identical, return the value
	- If both non-None and not identical, return None
	'''
	assert len(val_list) == 2

	if val_list[0] == val_list[1]: # both None or both non-None
		result = val_list[0]
	else:
		not_None_list = [ x != None for x in val_list ]
		if sum(not_None_list) == 1:
			result = next(itertools.compress(val_list, not_None_list))
		else: # both not None
			result = None

	return result


def get_seq_REFrange_from_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):

	if rpp.SV_supporting:
		seq_REFrange = None
	else:
		pos0_before_var, pos0_after_var = \
				module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)
	
		if check_containing_ambiguous_read(
			rpp, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var,
			pos0_after_var = pos0_after_var,
		):
			seq_REFrange = None
		else:
			seq_REFrange_list = list()
			for rp in (rpp.rp1, rpp.rp2):
				if 'seq_REFrange' not in dir(rp):
					seq_REFrange_list.append(
						module_getalleleclass_rp.get_seq_REFrange_from_read(
								rp.read, CHROM, POS0, REF, ALT,
								pairs_dict = rp.pairs_dict,
								pos0_before_var = pos0_before_var,
								pos0_after_var = pos0_after_var,
						)
					)
				else:
					seq_REFrange_list.append(rp.seq_REFrange)
		
			seq_REFrange = pick_from_two_values(seq_REFrange_list)

	return seq_REFrange


def get_alleleclass_from_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if rpp.SV_supporting:
		alleleclass = None
	else:
		pos0_before_var, pos0_after_var = \
				module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)
	
		if check_containing_ambiguous_read(
			rpp, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var,
			pos0_after_var = pos0_after_var,
		):
			alleleclass = None
		else:
			alleleclass_list = list()
			for rp in (rpp.rp1, rpp.rp2):
				if 'alleleclass' not in dir(rp):
					if 'seq_REFrange' not in dir(rp):
						alleleclass_list.append(
							module_getalleleclass_rp.get_alleleclass_from_read(
									rp.read, CHROM, POS0, REF, ALT, 
									pairs_dict = rp.pairs_dict,
									pos0_before_var = pos0_before_var,
									pos0_after_var = pos0_after_var,
									seq_REFrange = None,
							)
						)
					else:
						alleleclass_list.append(
							module_getalleleclass_rp.get_alleleclass_from_read(
									rp.read, CHROM, POS0, REF, ALT, 
									pairs_dict = rp.pairs_dict,
									pos0_before_var = pos0_before_var,
									pos0_after_var = pos0_after_var,
									seq_REFrange = rp.seq_REFrange,
							)
						)
				else:
					alleleclass_list.append(rp.alleleclass)

			alleleclass = pick_from_two_values(alleleclass_list)

	return alleleclass



import logging

import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup


DEFAULT_NUM_SUBMITTED = 400


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser()
    workflow.add_indir_arg(parser_dict['required'], 
                           required=True,
                           help=f'Directory where slurm job scripts exist.')
    workflow.add_checkint_arg(parser_dict['optional'], required=False)
    workflow.add_submitint_arg(parser_dict['optional'], required=False)
    workflow.add_logging_args(parser_dict)

    parser_dict['optional'].add_argument(
        '-n', required=False, default=DEFAULT_NUM_SUBMITTED, type=int, 
        help=(f'The number of jobs to submitted simultaneously. '
              f'Default: {DEFAULT_NUM_SUBMITTED}'))
    parser_dict['flag'].add_argument(
        '--nowait', action='store_true',
        help=(f'If set, this program will not wait for the submitted '
              f'jobs to finish.'))

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    jobscript_path_list = tools.listdir(args.indir_path)
    joblist = workflow.JobList(jobscript_path_list, logger=logger)
    joblist.submit(intv_submit=args.intv_submit)
    if not args.nowait:
        joblist.wait(intv_check=args.intv_check, edit_log_suffix=True)

    logger.info('ALL FINISHED SUCCESSFULLY')



import pandas as pd
import numpy as np
import os
import shutil
import warnings
import sys
warnings.filterwarnings("ignore")

def generateCNVMatrix(file_type, input_file, project, output_path, folder=False):

    def annotateSegFile(df, file_type, project, output_path):

        super_class = ['het', 'LOH', "homdel"]
        # het_sub_class = ['amp+', 'amp', 'gain', 'neut']
        # loh_subclass = ['amp+', 'amp', 'gain', 'neut', "del"]
        hom_del_class = ['0-100kb', '100kb-1Mb', '>1Mb']
        x_labels = ['>40Mb', '10Mb-40Mb', '1Mb-10Mb', '100kb-1Mb', '0-100kb']

        features = ['0:homdel:0-100kb', '0:homdel:100kb-1Mb', '0:homdel:>1Mb', '1:LOH:0-100kb',
            '1:LOH:100kb-1Mb', '1:LOH:1Mb-10Mb', '1:LOH:10Mb-40Mb', '1:LOH:>40Mb',
            '2:LOH:0-100kb', '2:LOH:100kb-1Mb', '2:LOH:1Mb-10Mb', '2:LOH:10Mb-40Mb', '2:LOH:>40Mb',
            '3-4:LOH:0-100kb', '3-4:LOH:100kb-1Mb', '3-4:LOH:1Mb-10Mb', '3-4:LOH:10Mb-40Mb', '3-4:LOH:>40Mb',
            '5-8:LOH:0-100kb', '5-8:LOH:100kb-1Mb', '5-8:LOH:1Mb-10Mb', '5-8:LOH:10Mb-40Mb', '5-8:LOH:>40Mb',
            '9+:LOH:0-100kb', '9+:LOH:100kb-1Mb', '9+:LOH:1Mb-10Mb', '9+:LOH:10Mb-40Mb', '9+:LOH:>40Mb',
            '2:het:0-100kb', '2:het:100kb-1Mb', '2:het:1Mb-10Mb', '2:het:10Mb-40Mb', '2:het:>40Mb',
            '3-4:het:0-100kb', '3-4:het:100kb-1Mb', '3-4:het:1Mb-10Mb', '3-4:het:10Mb-40Mb', '3-4:het:>40Mb',
            '5-8:het:0-100kb', '5-8:het:100kb-1Mb', '5-8:het:1Mb-10Mb', '5-8:het:10Mb-40Mb', '5-8:het:>40Mb',
            '9+:het:0-100kb', '9+:het:100kb-1Mb', '9+:het:1Mb-10Mb', '9+:het:10Mb-40Mb', '9+:het:>40Mb']

        assert(len(features) == 48)
        columns = list(df["sample"].unique())
        arr = np.zeros((48, len(columns)), dtype='int')
        nmf_matrix = pd.DataFrame(arr, index=features, columns=columns)


        # 2 - total copy number {del=0-1; neut=2; gain=3-4; amp=5-8; amp+=9+}
        CN_classes = ["1","2","3-4","5-8","9+"] # different total CN states
        CN_class = []
        if file_type == 'ASCAT_NGS':
            for tcn in df['Tumour TCN']:
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")

        elif file_type == 'SEQUENZA':
            for tcn in df['CNt']:
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == "ASCAT":
            for acn, bcn in zip(df['nMajor'], df['nMinor']):
                tcn = acn + bcn
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == 'ABSOLUTE':
            for acn, bcn in zip(df['Modal_HSCN_1'], df['Modal_HSCN_2']):
                tcn = acn + bcn
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == 'PCAWG':
            for tcn in df["copy_number"]:
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == 'FACETS':
            for tcn in df["tcn.em"]:
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == "BATTENBERG":
            for acn, bcn in zip(df['nMaj1_A'], df['nMin1_A']):
                tcn = int(acn) + int(bcn)
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        elif file_type == "PURPLE":
            for acn, bcn in zip(df['minorAlleleCopyNumber'], df['majorAlleleCopyNumber']):
                acn = int(acn)
                bcn = int(bcn)
                tcn = acn + bcn
                if tcn == 2:
                    CN_class.append("2")
                elif tcn == 0 or tcn == 1:
                    CN_class.append("1")
                elif tcn == 3 or tcn == 4:
                    CN_class.append("3-4")
                elif tcn >= 5 and tcn <= 8:
                    CN_class.append("5-8")
                else:
                    CN_class.append("9+")
        else:
            print("The file type " + str(file_type) + " is not currently supported. Please provide a proper file type.")

        df['CN_class'] = CN_class

        # 1 - LOH status {hom del; heterozygous; LOH}.
        LOH_status = []

        if file_type == 'ASCAT':
            for acn, bcn in zip(df['nMajor'], df['nMinor']):
                t = acn + bcn
                if t == 0:
                    LOH_status.append("homdel")
                elif acn == 0 or bcn == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")

        elif file_type == 'SEQUENZA':
            for t, a, b in zip(df['CNt'], df['A'], df['B']):
                if t == 0:
                    LOH_status.append("homdel")
                elif a == 0 or b == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")
        elif file_type == 'ASCAT_NGS':
            normal_ACN = np.asarray(df['Normal TCN']) - np.asarray(df['Normal BCN'])
            tumour_ACN = np.asarray(df['Tumour TCN']) - np.asarray(df['Tumour BCN'])
            df["Normal ACN"] = list(normal_ACN)
            df["Tumour ACN"] = list(tumour_ACN)

            A_CN = np.asarray(df['Tumour ACN']) #copy number of A allele in tumor
            B_CN = np.asarray(df['Tumour BCN']) #copy number of B allele in tumor
            loh = np.minimum(A_CN, B_CN) #minimum copy number when considering both A and B alleles
            df['loh'] = list(loh)
            for t, a in zip(df['Tumour TCN'], df['loh']):
                if t == 0:
                    LOH_status.append("homdel")
                elif a == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")

        elif file_type == 'ABSOLUTE':
            for acn, bcn in zip(df['Modal_HSCN_1'], df['Modal_HSCN_2']):
                t = acn + bcn
                if t == 0:
                    LOH_status.append("homdel")
                elif acn == 0 or bcn == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")
        elif file_type == 'PCAWG':
            for tcn, m in zip(df['copy_number'], df['mutation_type']):
                tcn = int(tcn)
                if m == 'copy neutral LOH' or m == 'amp LOH' or m == "hemizygous del LOH" or (m == "loss" and tcn == 1):
                    LOH_status.append("LOH")
                elif m == "copy neutral" or m == "gain"                                                      :
                    LOH_status.append("het")
                elif m == "loss" and tcn == 0:
                    LOH_status.append("homdel")
                else:
                    pass
                    #raise ValueError('Unable to determine zygosity')
        elif file_type == 'FACETS':
            for t, a in zip(df['tcn.em'], df['lcn.em']):
                if (t == 0 and a == 0) or (t == 0 and a == "NA") or (t == 0 and pd.isnull(a)):
                    LOH_status.append("homdel")
                elif t >= 1 and (a == 0 or a == "NA" or pd.isnull(a)):
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")
                    if t == 1:
                        print(t, a)

        elif file_type == "BATTENBERG":
            for acn, bcn in zip(df['nMaj1_A'], df['nMin1_A']):
                tcn = int(acn) + int(bcn)
                if tcn == 0:
                    LOH_status.append("homdel")
                elif acn == 0 or bcn == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")
        elif file_type == "PURPLE":
            for acn, bcn in zip(df['minorAlleleCopyNumber'], df['majorAlleleCopyNumber']):
                acn = int(acn)
                bcn = int(bcn)
                tcn = acn + bcn
                #print(acn,bcn, tcn)
                if tcn == 0:
                    LOH_status.append("homdel")
                elif acn == 0 or bcn == 0:
                    LOH_status.append("LOH")
                else:
                    LOH_status.append("het")

        else:
            print("The file type " + str(file_type) + " is not currently supported. Please provide a proper file type.")



        df['LOH'] = LOH_status

        lengths = []

        #get chromosomal sizes
        if file_type == 'ASCAT_NGS':
            for start, end in zip(df['Start Position'], df['End Position']):
                lengths.append((end - start)/1000000) #megabases
        elif file_type == 'SEQUENZA':
            for start, end in zip(df['start.pos'], df['end.pos']):
                lengths.append((end - start)/1000000)
        elif file_type == 'ABSOLUTE': #Start End
            for start, end in zip(df['Start'], df['End']):
                lengths.append((end - start)/1000000)
        elif file_type == 'ASCAT':
            for start, end in zip(df['startpos'], df['endpos']):
                lengths.append((end - start)/1000000)
        elif file_type == 'PCAWG':
            for start, end in zip(df['chromosome_start'], df['chromosome_end']):
                lengths.append((end - start)/1000000)
        elif file_type == 'FACETS' or file_type == "PURPLE":
            for start, end in zip(df['start'], df['end']):
                lengths.append((end - start)/1000000)
        elif file_type == "BATTENBERG":
            for start, end in zip(df['startpos'], df['endpos']):
                lengths.append((end - start)/1000000)
        else:
             print("The file type " + str(file_type) + " is not currently supported. Please provide a proper file type.")


        df['length'] = lengths

        sizes = []
        size_bins = []
        hom_del_class = ['0-100kb', '100kb-1Mb', '>1Mb']

        for l, s in zip(lengths, df['LOH']): #keep in mind the lengths are in megabases
            if s == 'homdel':
                if l > -0.01 and l <= 0.1:
                    size = "0-100kb"
                    size_bin = "(-0.01,0.1]"
                elif l > 0.1 and l <= 1:
                    size = "100kb-1Mb"
                    size_bin = "(0.1,1]"
                else:
                    size = '>1Mb'
                    size_bin = "(1,Inf]"
            else:
                if l > -0.01 and l <= 0.1:
                    size = "0-100kb"
                    size_bin = "(-0.01,0.1]"
                elif l > 0.1 and l <= 1:
                    size = "100kb-1Mb"
                    size_bin = "(0.1,1]"
                elif l > 1 and l <= 10:
                    size = "1Mb-10Mb"
                    size_bin = "(1,10]"
                elif l > 10 and l <= 40:
                    size = "10Mb-40Mb"
                    size_bin = "(10,40]"
                else:
                    size = ">40Mb"
                    size_bin = "(40,Inf]"
            sizes.append(size)
            size_bins.append(size_bin)
        df['size_classification'] = sizes

        for sample, tcn, loh, size in zip(df["sample"], df['CN_class'], df['LOH'], df['size_classification']):
            if loh == "homdel":
                channel = "0" + ":" + loh + ":" + size
            else:
                channel = tcn + ":" + loh + ":" + size
            nmf_matrix.at[channel, sample] += 1

        nmf_matrix.index.name = 'MutationType'
        # output_path = output_path + project + "/"
        # if os.path.exists(output_path):
        #     shutil.rmtree(output_path)
        # os.makedirs(output_path)
        nmf_matrix.reset_index(level=0, inplace=True)
        if file_type != "BATTENBERG":
            out_file_path=os.path.join(output_path, project + '.CNV48.matrix.tsv')
            nmf_matrix.to_csv(out_file_path, sep='\t', index=None)
            print("Saved matrix to " + out_file_path)
        return nmf_matrix, df


    # create output_dir if it does not yet exist
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    df = pd.read_csv(input_file, sep='\t')
    if file_type == "BATTENBERG":
        clonal_df = df[["sample", "chr", "startpos", "endpos", "nMaj1_A", "nMin1_A"]]
        subclonal_df = df[["sample", "chr", "startpos", "endpos", "nMaj2_A", "nMin2_A"]]
        clonal_df = clonal_df.dropna()
        subclonal_df = subclonal_df.dropna()
        subclonal_df = subclonal_df.rename(columns={"nMaj2_A": "nMaj1_A", "nMin2_A": "nMin1_A"})
        clonal_matrix = annotateSegFile(clonal_df, file_type, project, output_path)[0]
        subclonal_matrix = annotateSegFile(subclonal_df, file_type, project, output_path)[0]
        subclonal_matrix = subclonal_matrix.set_index(subclonal_matrix.columns[0])
        clonal_matrix = clonal_matrix.set_index(clonal_matrix.columns[0])
        #add subclonal events to matrix containing clonal events to produce a final matrix that accounts for all events(Clonal + Subclonal)
        for k in range(0,len(subclonal_matrix)):
            for i,j in enumerate(subclonal_matrix.columns):
                clonal_matrix.at[subclonal_matrix.index.to_list()[k], j] = subclonal_matrix.iat[k, i] + clonal_matrix.loc[subclonal_matrix.index.to_list()[k], j]
        out_file_path = os.path.join(output_path, project + '.CNV48.matrix.tsv')
        clonal_matrix.to_csv(out_file_path, sep='\t')
        nmf_matrix = clonal_matrix
        print("Saved matrix to " + out_file_path)
    elif file_type == "VCF":
        #convert vcf to dataframe with necessary info
        vcf_file = vcf.Reader(open(file, 'r'))
        sample_name=file.split(".vcf")[0]
        # include the file-header:
        b = []
        l = []
        c = []
        samples = []
        starts= []
        ends = []
        for record in vcf_file:
            if record.INFO.get('LCN_EM') is not None and record.INFO.get('TCN_EM') is not None:
                tcn = record.INFO.get('TCN_EM')
                lcn = record.INFO.get('LCN_EM')
                bcn = abs(tcn-lcn)
                b.append(bcn)
                l.append(lcn)
                assert(tcn >= lcn)
                samples.append(sample_name)
                starts.append(record.POS)
                ends.append(record.INFO.get('END'))
                c.append(record.CHROM)
        df = pd.Dataframe({"sample":samples, "chr":c, "startpos":starts, "endpos":ends, "nMajor":tcn, "nMinor":lcn})
        nmf_matrix, annotated_df = annotateSegFile(df, "ASCAT", project, output_path)
    else:
        nmf_matrix, annotated_df = annotateSegFile(df, file_type, project, output_path)

    return nmf_matrix

if __name__ == "__main__":
    if len(sys.argv) > 1:
        file_type, input_dir, project, output_path = sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4]
        generateCNVMatrix(file_type, input_dir, project, output_path)



import sys
import os
import subprocess
import pathlib
import importlib
import importlib.resources
import importlib.metadata

import yaml

import handygenome


__version__ = importlib.metadata.version(handygenome.__name__)


#####################################################################################################
# set numpy threads (https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy) #
#####################################################################################################

os.environ["OMP_NUM_THREADS"] = "1"  # export OMP_NUM_THREADS=4
os.environ["OPENBLAS_NUM_THREADS"] = "1"  # export OPENBLAS_NUM_THREADS=4 
os.environ["MKL_NUM_THREADS"] = "1"  # export MKL_NUM_THREADS=6
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"  # export VECLIB_MAXIMUM_THREADS=4
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # export NUMEXPR_NUM_THREADS=6


############################
# USERDIR (~/.handygenome) #
############################

USERDIR = pathlib.Path.home() / '.handygenome'

USERDATA_DIR = USERDIR / 'data'
CONFIGPATH = USERDIR / 'config.yaml'


#####################
# CONFIG and PARAMS #
#####################

with CONFIGPATH.open() as f:
    OPTION = yaml.safe_load(f)


def get_exec_path(key, replace=None, whicharg=None):
    if OPTION[key] is None:
        if replace is None:
            if whicharg is None:
                whicharg = key

            p = subprocess.run(['which', whicharg], capture_output=True, text=True)
            if p.returncode == 0:
                return p.stdout.strip()
            else:
                return None
        else:
            return replace
    else:
        return OPTION[key]


PARAMS = dict()
PARAMS['python'] = get_exec_path('python', replace=sys.executable)
PARAMS['bash'] = get_exec_path('bash')
PARAMS['perl'] = get_exec_path('perl')
PARAMS['mosdepth'] = get_exec_path('mosdepth')
PARAMS['default_refver'] = OPTION['default_refver']


#################################
# data files within top package #
#################################

PKGPATH = importlib.resources.files(handygenome)
DATA_DIR = PKGPATH.joinpath('data')

DIRS = {
    #'data': os.path.join(PROJECT_PATH, 'data'),
    #'externals': os.path.join(PROJECT_PATH, 'externals'),
    'R': PKGPATH.joinpath('Rscripts'),
    'scripts': PKGPATH.joinpath('scripts'),
    #'package': os.path.join(PROJECT_PATH, 'handygenome'),
}



import os
import re
import argparse
import gzip
import ftplib
import shutil

import pysam

import handygenome.network as network
import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.annotation.customfile as customfile
import handygenome.refgenome.refgenome as refgenome


DEFAULT_OUTFILE_BASENAME = 'ensembl_regulatory_sorted.gff.gz'

MAINFILE_URL_GRCH37 = 'http://ftp.ensembl.org/pub/grch37/release-105/regulation/homo_sapiens/homo_sapiens.GRCh37.Regulatory_Build.regulatory_features.20201218.gff.gz'
MAINFILE_URL_GRCH38 = 'http://ftp.ensembl.org/pub/current_regulation/homo_sapiens/homo_sapiens.GRCh38.Regulatory_Build.regulatory_features.20210107.gff.gz'

FTP_HOST = 'ftp.ensembl.org'
FTP_ACTIVITY_DIRECTORY_GRCH37 = '/pub/grch37/release-105/regulation/homo_sapiens/RegulatoryFeatureActivity'
FTP_ACTIVITY_DIRECTORY_GRCH38 = '/pub/current_regulation/homo_sapiens/RegulatoryFeatureActivity'

DEFAULT_OUTFILE_PATH_GRCH37 = os.path.join(
    os.getcwd(), re.sub('gff.gz$', 'grch37.gff.gz', DEFAULT_OUTFILE_BASENAME))
DEFAULT_OUTFILE_PATH_GRCH38 = os.path.join(
    os.getcwd(), re.sub('gff.gz$', 'grch38.gff.gz', DEFAULT_OUTFILE_BASENAME))


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser(
        description=(f'Ensembl gff files for regulatory elements and '
                     f'activity are downloaded and merged into a single '
                     f'gff file.'))

    parser_dict['optional'].add_argument(
        '--outfile-grch37', dest='outfile_path_grch37', 
        default=DEFAULT_OUTFILE_PATH_GRCH37,
        type=workflow.arghandler_outfile_ask,
        required=False, help=f'Output gff file path for grch37')

    parser_dict['optional'].add_argument(
        '--outfile-grch38', dest='outfile_path_grch38', 
        default=DEFAULT_OUTFILE_PATH_GRCH38,
        type=workflow.arghandler_outfile_ask,
        required=False, help=f'Output gff file path for grch38')

    parser_dict['optional'].add_argument(
        '--download-dir-grch37', dest='download_dir_grch37', required=False, 
        help=f'Directory of downloaded files for grch37. Not removed.')

    parser_dict['optional'].add_argument(
        '--download-dir-grch38', dest='download_dir_grch38', required=False, 
        help=f'Directory of downloaded files for grch38. Not removed.')

    workflow.add_logging_args(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def download(logger, mainfile_url, ftp_activity_directory, download_dir, 
             mainfile_path, activity_dir):
    # main file
    logger.info(f'Beginning download of main gff to "{mainfile_path}"')
    network.download(mainfile_url, mainfile_path)

    # activity files
    logger.info(f'LOGIN to ensembl ftp host "{FTP_HOST}"')
    ftp = ftplib.FTP(FTP_HOST)
    ftp.login()
    ftp.cwd(ftp_activity_directory)
    tissue_list = ftp.nlst()
    for tissue in tissue_list:
        logger.info(f'Beginning download of activity file for "{tissue}"')
        local_fname = os.path.join(activity_dir, f'{tissue}.gff.gz')
        ftp.cwd(tissue)
        fname = [x for x in ftp.nlst() if x.startswith('homo_sapiens')][0]
        with open(local_fname, 'wb') as outfile:
            ftp.retrbinary(f'RETR {fname}', outfile.write)
        ftp.cwd('..')

    ftp.quit()

    logger.info(f'Download finished')


def load_activity_data(activity_dir, logger):
    activity_data = dict()
    tissue_list = list()
    for fname in tools.listdir(activity_dir):
        tissue = re.sub('\.gff\.gz$', '', os.path.basename(fname))
        tissue_list.append(tissue)
        logger.info(f'Loading {tissue} file')
        with gzip.open(fname, 'rt') as infile:
            for line in infile:
                linesp = tools.get_linesp(line)
                attrs = linesp[8]
                attrs_parsed = customfile.parse_gff3_attrs(attrs)
                ID = attrs_parsed['regulatory_feature_stable_id']
                activity = attrs_parsed['activity']

                activity_data.setdefault(ID, dict())
                activity_data[ID][tissue] = activity

    return activity_data, tissue_list


def load_mainfile(mainfile_path, chromdict, asmblspec, 
                  output_chrname_version):
    mainfile_data = list()
    with gzip.open(mainfile_path, 'rt') as infile:
        for line in infile:
            linesp = tools.get_linesp(line)

            # edit chrom
            if linesp[0] not in chromdict:
                linesp[0] = asmblspec.convert(linesp[0], output_chrname_version)

            # get attributes dictionary
            raw_attrs = customfile.parse_gff3_attrs(linesp[8])
            attrs = {
                'id': raw_attrs['ID'].split(':')[1],
                'bound_end': raw_attrs['bound_end'],
                'bound_start': raw_attrs['bound_start'],
            }

            # remove linesp attributes field
            del linesp[8]

            line_data = {'linesp': linesp, 'attrs': attrs}
            mainfile_data.append(line_data)

    return mainfile_data


def sort_mainfile_data(chromdict, mainfile_data):
    def sortkey(line_data):
        return tools.coord_sortkey(line_data['linesp'][0], int(line_data['linesp'][3]), chromdict)
    mainfile_data_sorted = sorted(mainfile_data, key=sortkey)

    return mainfile_data_sorted


def write_uncompressed_file(outfile_path_uncomp, mainfile_data_sorted, activity_data, tissue_list):
    with open(outfile_path_uncomp, 'wt') as outfile:
        for line_data in mainfile_data_sorted:
            activity = activity_data[line_data['attrs']['id']]
            line_data['attrs']['activity'] = ','.join(
                f'{tissue}:{activity.setdefault(tissue, "NA")}' 
                for tissue in tissue_list)
            attrs_string = ';'.join(
                f'{key}={val}' for key, val in line_data['attrs'].items())
            line_data['linesp'].append(attrs_string)
            outfile.write('\t'.join(line_data['linesp']) + '\n')


def convert_common(logger, mainfile_url, ftp_activity_directory, refver, 
                   output_chrname_version, outfile_path, download_dir):
    # download
    if download_dir is None:
        download_dir = workflow.get_tmpfile_path(delete=False, is_dir=True)
        mainfile_path = os.path.join(download_dir, 'main.gff.gz')
        activity_dir = os.path.join(download_dir, 'activity')
        os.mkdir(activity_dir)
        download(logger, mainfile_url, ftp_activity_directory, download_dir, 
                 mainfile_path, activity_dir)

        rm_downloaded = True
    else:
        mainfile_path = os.path.join(download_dir, 'main.gff.gz')
        activity_dir = os.path.join(download_dir, 'activity')

        rm_downloaded = False

    # set refgenome dbs
    asmblspec = refgenome.SPECS[refver]
    chromdict = asmblspec.chromdicts[output_chrname_version]

    # load activity data
    logger.info(f'Loading activity files')
    activity_data, tissue_list = load_activity_data(activity_dir, logger)

    # load main file
    logger.info(f'Loading the main file')
    mainfile_data = load_mainfile(mainfile_path, chromdict, asmblspec, 
                                  output_chrname_version)

    # sort
    logger.info(f'Sorting loaded main file data')
    mainfile_data_sorted = sort_mainfile_data(chromdict, mainfile_data)

    # write uncompressed outfile
    logger.info(f'Writing the sorted lines')
    outfile_path_uncomp = outfile_path + '.uncompressed'
    write_uncompressed_file(outfile_path_uncomp, mainfile_data_sorted, 
                            activity_data, tissue_list)

    # compress and index
    logger.info(f'Compressing with bgzip and indexing with tabix')
    pysam.tabix_compress(outfile_path_uncomp, outfile_path)
    pysam.tabix_index(outfile_path, preset = 'gff')

    # remove temp files
    os.remove(outfile_path_uncomp)
    if rm_downloaded:
        shutil.rmtree(download_dir)


def convert_grch37(logger, outfile_path, download_dir):
    convert_common(logger, MAINFILE_URL_GRCH37, FTP_ACTIVITY_DIRECTORY_GRCH37, 'hg19', 'nochr_plus_genbank', outfile_path, download_dir)


def convert_grch38(logger, outfile_path, download_dir):
    convert_common(logger, MAINFILE_URL_GRCH38, FTP_ACTIVITY_DIRECTORY_GRCH38, 'hg38', 'ucsc', outfile_path, download_dir)


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    #convert_grch37(logger, args.outfile_path_grch37, args.download_dir_grch37)
    convert_grch38(logger, args.outfile_path_grch38, args.download_dir_grch38)
    
    logger.info('ALL SUCCESSFULLY FINISHED')



import os
import re
import inspect
import textwrap
import stat
import string

import pysam
import pyranges as pr

import handygenome
import handygenome.tools as tools
import handygenome.interval as libinterval
import handygenome.workflow as workflow
import handygenome.vcfeditor.misc as vcfmisc
import handygenome.ucscdata as ucscdata
import handygenome.blacklist as blacklist


SPLIT_INFILE_PAT = re.compile(r'([0-9]+)([A-Za-z]*)(\.vcf\.gz)')
LETTERS = sorted(string.ascii_letters)


def make_next_split_vcf_path(fname):
    """Examples:
        012.vcf.gz => 012A.vcf.gz => 012B.vcf.gz => ...
        0114Z.vcf.gz => 0114a.vcf.gz => 0114b.vcf.gz => ...
        0234z.vcf.gz => 0234zA.vcf.gz => 0234zB.vcf.gz => ...
    """
    bname = os.path.basename(fname)
    dname = os.path.dirname(fname)
    mat = SPLIT_INFILE_PAT.fullmatch(bname)
    assert mat is not None

    chars = mat.group(2)
    if chars == '':
        next_chars = LETTERS[0]
    else:
        lastchar = chars[-1]
        if lastchar == LETTERS[-1]:
            next_chars = chars + LETTERS[0]
        else:
            next_chars = chars[:-1] + LETTERS[LETTERS.index(lastchar) + 1]

    next_bname = mat.group(1) + next_chars + mat.group(3)
    return os.path.join(dname, next_bname)

make_next_infile = make_next_split_vcf_path


def setup_logger(args, logger_name=None, tmpdir_root=None, 
                 with_genlog=False):
    if logger_name is None:
        caller_frame = inspect.stack()[1].frame
        caller_module = inspect.getmodule(caller_frame)
        logger_name = caller_module.__name__.split('.')[-1]

    filenames = list()
    if args.log is not None:
        filenames.append(args.log)
    if with_genlog:
        genlog_path = os.path.join(tmpdir_root, 'genlog.txt')
        filenames.append(genlog_path)

    logger = workflow.get_logger(name=logger_name, 
                                 stderr=(not args.silent),
                                 filenames=filenames, 
                                 append=args.append_log)

    return logger


def handle_region_args(chromdict, incl_bed_path, excl_bed_path, num_split, regionfiles_dir):
    incl_intvlist = get_included_intvlist(chromdict, incl_bed_path, excl_bed_path)
    incl_intvlist_split = incl_intvlist.split(num=num_split)
    # write
    for zidx, intvlist in tools.zenumerate(incl_intvlist_split):
        fname = os.path.join(regionfiles_dir, f"{zidx}.bed")
        intvlist.write_bed(fname)
    # make padded versions and write
    for intvlist in incl_intvlist_split:
        intvlist.slop(chromdict, b=5000)
    for zidx, intvlist in tools.zenumerate(incl_intvlist_split):
        fname = os.path.join(regionfiles_dir, f"{zidx}.padded.bed")
        intvlist.write_bed(fname)


def get_included_intvlist(chromdict, incl_bed_path, excl_bed_path):
    if incl_bed_path is None:
        incl_intvlist = libinterval.IntervalList.from_chromdict(chromdict)
    else:
        incl_intvlist = libinterval.IntervalList.from_bed(incl_bed_path)

    if excl_bed_path is not None:
        excl_intvlist = libinterval.IntervalList.from_bed(excl_bed_path)
        incl_intvlist = incl_intvlist.subtract(excl_intvlist)

    incl_intvlist.sort_intervals(chromdict)

    return incl_intvlist


def write_region_files(regionfiles_dir, incl_intvlist_split, incl_intvlist_split_padded):
    for zidx, intvlist in tools.zenumerate(incl_intvlist_split):
        fname = os.path.join(regionfiles_dir, f"{zidx}.bed")
        intvlist.write_bed(fname)

    for zidx, intvlist in tools.zenumerate(incl_intvlist_split_padded):
        fname = os.path.join(regionfiles_dir, f"{zidx}.padded.bed")
        intvlist.write_bed(fname)


def write_jobscripts(
    script_path_list,
    log_path_list,
    slurm_log_pf_list,
    module_name,
    unit_job_func_name,
    kwargs_single,
    kwargs_multi,
    jobname_prefix,
    nproc,
):
    # sanity check
    arglens = list()
    arglens.append(len(script_path_list))
    arglens.append(len(log_path_list))
    arglens.extend(len(val) for val in kwargs_multi.values())
    if len(set(arglens)) != 1:
        raise Exception(
            f'The lengths of "script_path_list", "log_path_list", '
            f'and components of "kwargs_multi" are not all the same.'
        )
    for key, val in kwargs_single.items():
        if val != eval(repr(val)):
            raise Exception(f'Keyword argument {key}: {val} cannot be represented as string')
    for key, vallist in kwargs_multi.items():
        for val in vallist:
            if val != eval(repr(val)):
                raise Exception(f'Keyword argument {key}: {val} cannot be represented as string')

    # main
    tab = ' ' * 4
    for zidx, (script_path, log_path, slurm_log_pf) in tools.zenumerate(
        zip(script_path_list, log_path_list, slurm_log_pf_list)
    ):
        idx = tools.zidx_to_idx(zidx)
        success_log_path = re.sub("\.log$", ".success", log_path)
        failure_log_path = re.sub("\.log$", ".failure", log_path)
        slurmlog_out_path = slurm_log_pf + '.out'
        slurmlog_err_path = slurm_log_pf + '.err'

        unit_job_func_args = list()
        for key, val in kwargs_single.items():
            unit_job_func_args.append(f"{key}={repr(val)},")
        for key, val in kwargs_multi.items():
            unit_job_func_args.append(f"{key}={repr(val[idx])},")
        unit_job_func_args = ('\n' + (7 * tab)).join(unit_job_func_args)

        script_contents = list()
        script_contents.append(
            textwrap.dedent(f"""\
                #!{handygenome.OPTION["python"]}

                #SBATCH -N 1
                #SBATCH -n 1

                #SBATCH -c {nproc}
                #SBATCH -o {slurmlog_out_path}
                #SBATCH -e {slurmlog_err_path}
                #SBATCH -J {jobname_prefix}_{zidx}

                import os
                import contextlib
                import traceback
                import sys
                sys.path.append({repr(handygenome.PROJECT_PATH)})
                from {module_name} import {unit_job_func_name}

                log = open({repr(log_path)}, 'w')
                with contextlib.redirect_stdout(log), contextlib.redirect_stderr(log):
                    try:
                        {unit_job_func_name}(
                            {unit_job_func_args}
                        )
                    except:
                        print(traceback.format_exc())
                        success = False
                    else:
                        success = True
                log.close()

                if success:
                    os.rename({repr(log_path)}, {repr(success_log_path)})
                else:
                    os.rename({repr(log_path)}, {repr(failure_log_path)})
                    raise SystemExit(1)"""
            )
        )

        script_contents = '\n'.join(script_contents)
        with open(script_path, "w") as outfile:
            outfile.write(script_contents)

        os.chmod(script_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)


def get_outfile_script_log_paths(split_infile_path_list, outfile_dir, script_dir, log_dir):
    split_outfile_path_list = list()
    script_path_list = list()
    log_path_list = list()

    for zidx, split_infile_path in tools.zenumerate(split_infile_path_list):
        basename = os.path.basename(split_infile_path)
        split_outfile_path_list.append(
            os.path.join(outfile_dir, basename)
        )
        script_path_list.append(
            os.path.join(script_dir, f"{zidx}.sbatch")
        )
        log_path_list.append(
            os.path.join(log_dir, f"{zidx}.sbatch.log")
        )

    return split_outfile_path_list, script_path_list, log_path_list


def get_script_log_paths(script_dir, log_dir, num_split):
    script_path_list = list()
    log_path_list = list()

    for zidx in tools.zrange(num_split):
        script_path_list.append(os.path.join(script_dir, f"{zidx}.sbatch"))
        log_path_list.append(os.path.join(log_dir, f"{zidx}.sbatch.log"))

    return script_path_list, log_path_list


def make_infile_copy(infile_path, tmpdir_root, logger):
    infile_link_path = os.path.join(tmpdir_root, 'infile_link.vcf.gz')
    is_vcf, comp, is_bgzf = vcfmisc.get_vcf_format(infile_path)
    if is_bgzf:
        os.symlink(infile_path, infile_link_path)
        indexfile_path = vcfmisc.get_indexfile_path(infile_path)
        if indexfile_path is None:
            logger.info(
                f'Making index of input VCF (which will be saved in the temporary directory)'
            )
            infile_link_index_path = vcfmisc.make_index(infile_link_path)
        else:
            infile_link_index_path = f'{infile_link_path}.csi'
            os.symlink(indexfile_path, infile_link_index_path)
    else:
        logger.info(f'Making bgzipped copy of input VCF')
        in_vcf = pysam.VariantFile(infile_path)
        out_vcf = pysam.VariantFile(infile_link_path, mode='wz', header=in_vcf.header.copy())
        for vr in in_vcf.fetch():
            out_vcf.write(vr)

        out_vcf.close()
        in_vcf.close()

        logger.info(f'Making index of the copied VCF')
        infile_link_index_path = vcfmisc.make_index(infile_link_path)

    return infile_link_path, infile_link_index_path
            
        
def get_blacklist_gr(refver):
    cytoband_gr = ucscdata.get_cytoband(refver=refver, as_gr=True)
    centromere_gr = cytoband_gr[cytoband_gr.Stain == 'acen']
    blacklist_gr = pr.concat([centromere_gr, blacklist.HIGHDEPTH_BLACKLIST_GR])
    blacklist_gr = blacklist_gr.sort().merge()

    return blacklist_gr




import importlib
top_package_name = __name__.split('.')[0]
getalleleclass_rp = importlib.import_module('.'.join([top_package_name, 'ria', 'get_alleleclass_from_readplus']))
readhandler = importlib.import_module('.'.join([top_package_name, 'readplus', 'readhandler']))
common = importlib.import_module('.'.join([top_package_name, 'common']))


# seq_REFrange and alleleclass

def add_seq_REFrange_to_readplus(
		rp, CHROM, POS0, REF, ALT,
		pos0_before_var = None,
		pos0_after_var = None,
		):

	pos0_before_var, pos0_after_var = \
			getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	rp.seq_REFrange = getalleleclass_rp.get_seq_REFrange_from_read(
		rp.read, CHROM, POS0, REF, ALT, 
		pairs_dict = rp.pairs_dict,
		pos0_before_var = pos0_before_var, 
		pos0_after_var = pos0_after_var,
		)


def add_alleleclass_to_readplus(
		rp, CHROM, POS0, REF, ALT, 
		pos0_before_var = None, 
		pos0_after_var = None,
		seq_REFrange = None,
		):

	pos0_before_var, pos0_after_var = \
			getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	rp.alleleclass = getalleleclass_rp.get_alleleclass_from_read(
		rp.read, CHROM, POS0, REF, ALT, 
		pairs_dict = rp.pairs_dict, 
		pos0_before_var = pos0_before_var, 
		pos0_after_var = pos0_after_var,
		seq_REFrange = seq_REFrange,
		)


def add_seqREFrange_alleleclass_to_readplus(
		rp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):

	pos0_before_var, pos0_after_var = \
			getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	add_seq_REFrange_to_readplus(
			rp, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var,
			pos0_after_var = pos0_after_var,
			)

	add_alleleclass_to_readplus(
		rp, CHROM, POS0, REF, ALT, 
		pos0_before_var = pos0_before_var,
		pos0_after_var = pos0_after_var,
		seq_REFrange = rp.seq_REFrange,
		)


# other attributes

def add_IDCinfo_to_readplus(rp):
	rp.len_ins = rp.cigarstats[1]
	rp.len_del = rp.cigarstats[2]
	rp.len_clip = rp.cigarstats[4]


def add_MM_to_readplus(rp, REF, ALT, mttype=None):
	'''
- MM: Stands for mismatch. Represents only the number of mismatches with cigar M operation.
- NM tag value includes indels as well as base mismatch ; indel length is subtracted
- Works even if the read does not have NM tag
	'''
	if mttype == None:
		mttype = common.get_mttype(REF, ALT)

	if rp.read.has_tag('NM'):
		rp.MM = rp.read.get_tag('NM') - rp.cigarstats[1] - rp.cigarstats[2]
	else:
		rp.MM = readhandler.get_MM(rp.read, rp.pairs_dict)

	# subtract the number of mismatch attributed to the variant itself
	if mttype == 'snv' or mttype == 'mnv':
		rp.MM -= len(REF)


def add_others_to_readplus(rp):
	rp.lowqual = False
	rp.lowqual_cause = list()
	rp.SVread = False
	rp.SVread_cause = list()


# main function

def ria_modify_readplus(
		rp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if rp.irrelevant:
		raise Exception('Irrelevant readplus object was entered.')

	add_seqREFrange_alleleclass_to_readplus(rp, CHROM, POS0, REF, ALT)
	add_IDCinfo_to_readplus(rp)
	add_MM_to_readplus(rp, REF, ALT)
	add_others_to_readplus(rp)








import copy

import pysam

import handygenome.annotation.annotitem as libannotitem
import handygenome.annotation.customfile as customfile


class CosmicInfo(libannotitem.AnnotItemVariantInfoSingle):
    @classmethod
    def init_blank(cls, metadata=None):
        result = cls.init_nonmissing()
        result['id'] = None
        result['occurrence'] = None
        result['occurrence_somatic'] = None
        result['coding_score'] = None
        result['noncoding_score'] = None
        result.metadata = metadata
        return result
    ##############

    def get_self_show(self):
        self_show = dict()

        for key in (
            'id', 
            'occurrence', 'portion_by_site', 'total_occurrence', 'total_portion', 
            'occurrence_somatic', 'portion_by_site_somatic', 'total_occurrence_somatic', 'total_portion_somatic', 
        ):
            try:
                self_show[key] = self[key]
            except KeyError:
                self_show[key] = None

        for key in ('occurrence', 'occurrence_somatic',
                    'portion_by_site', 'portion_by_site_somatic'):
            try:
                if self_show[key] is not None:
                    self_show[key] = dict(sorted(self_show[key].items(),
                                                 key=(lambda x: x[1]),
                                                 reverse=True))
            except KeyError:
                self_show[key] = None

        return self_show

    def __getitem__(self, key):
        """Real dictionary keys are: "id", "occurrence", "occurrence_somatic", 
        "coding_score", "noncoding_score"
        """
        if key == 'portion_by_site':
            return self.get_portion_by_site(somatic=False)
        elif key == 'portion_by_site_somatic':
            return self.get_portion_by_site(somatic=True)
        elif key == 'total_occurrence':
            return self.get_total_occurrence(somatic=False)
        elif key == 'total_occurrence_somatic':
            return self.get_total_occurrence(somatic=True)
        elif key == 'total_portion':
            return self.get_total_portion(somatic=False)
        elif key == 'total_portion_somatic':
            return self.get_total_portion(somatic=True)
        else:
            return super().__getitem__(key)

    def get_portion_by_site(self, somatic=False):
        key = 'occurrence_somatic' if somatic else 'occurrence'
        if self[key] is None:
            return None
        else:
            portion_by_site = dict()
            for site, count in self[key].items():
                num_sample_site = self.metadata['num_sample_by_site'][site]
                frac = count / num_sample_site
                portion_by_site[site] = frac
            return portion_by_site

    def get_total_occurrence(self, somatic=False):
        key = 'occurrence_somatic' if somatic else 'occurrence'
        if self[key] is None:
            return None
        else:
            return sum(self[key].values())

    def get_total_portion(self, somatic=False):
        total_occurrence = self.get_total_occurrence(somatic=somatic)
        if total_occurrence is None:
            return None
        else:
            num_sample_allsites = sum(self.metadata['num_sample_by_site'].values())
            return total_occurrence / num_sample_allsites


class CosmicMetadata(libannotitem.AnnotItemHeader):
    meta = {'ID': 'cosmic_metadata'}

    @classmethod
    def from_vcfheader(cls, vcfheader):
        return cls.from_vcfheader_base(vcfheader)

    def write(self, vcfheader):
        self.write_base(vcfheader)


class CosmicInfoALTlist(libannotitem.AnnotItemVariantInfoALTlist):
    meta = {
        'ID': 'cosmic_info', 
        'Number': 'A', 
        'Type': 'String', 
        'Description': 'COSMIC information encoded as a string, one for each ALT allele',
    }

    unit_class = CosmicInfo
    metadata_class = CosmicMetadata

    @classmethod
    def from_vr(cls, vr, metadata=None):
        if metadata is None:
            metadata = cls.metadata_class.from_vcfheader(vr.header)

        result = cls.from_vr_base(vr)
        for unit in result:
            unit.metadata = metadata

        return result

    @classmethod
    def from_vcfspec(cls, vcfspec, cosmic_vcf, metadata=None, donot_init_metadata=False):
        if metadata is None:
            if not donot_init_metadata:
                metadata = cls.metadata_class.from_vcfheader(cosmic_vcf.header)

        cosmic_vr_list = customfile.fetch_relevant_vr_multialt(
            vcfspec, cosmic_vcf, search_equivs=True, raise_with_multihit=True,
        )
        result = cls()
        for vr in cosmic_vr_list:
            if vr is None:
                result.append(cls.unit_class.init_blank(metadata=metadata))
                #result.append(cls.unit_class.init_missing())
            else:
                result.extend(cls.from_vr(vr, metadata=metadata))

        return result

    def write(self, vr):
        self.write_base(vr)


def update_vcfheader(vcfheader, cosmic_vcf):
    CosmicInfoALTlist.add_meta(vcfheader)
    cosmicmeta = CosmicMetadata.from_vcfheader(cosmic_vcf.header)
    cosmicmeta.write(vcfheader)


#def annotate_vp(vp, cosmic_vcf, fasta, metadata=None):
#    vp.cosmic = CosmicInfoALTlist.from_vcfspec(vp.vcfspec, cosmic_vcf, fasta, metadata=metadata)





import itertools

import Bio.Align

import handygenome.annotation.annotitem as annotitem


class GenotypeInfo(annotitem.AnnotItemInfoSingle):
    @classmethod
    def from_vp(
        cls, vp, sampleinfo, 
        germline_CNt=2,
        cutoff_somatic_vaf={
            'blood': 0.05, 'tumor': 0.05, 'clonal': 0.2,
        },
    ):
        """Args:
            sampleinfo: 
                format: dict(
                    sampleid=dict(
                        "type"="clonal"/"blood"/"tumor", 
                        "CNt"=<int>,
                    )
                )
                Assumes all samples are derived from the same individual
        """
        pass
        # set params
        possible_allele_indexes = set(range(len(vp.vcfspec.alts) + 1))
        sampletypes = {'clonal': list(), 'blood': list(), 'tumor': list()}
        for sampleid, subdic in sampleinfo.items():
            sampletypes[subdic['type']].append(sampleid)

        if (
            len(sampletypes['clonal']) == 0 and 
            len(sampletypes['blood']) == 0
        ):
            raise Exception(f'No blood or clonal samples; cannot determine germline genotype')

        # determine germline
        allele_index_polls = dict()
        for sampleid in itertools.chain(sampletypes['clonal'], sampletypes['blood']):
            rppcount_dict = vp.readstats_dict[sampleid]['rppcounts']
            allele_rppcounts_sorted = sorted(
                (
                    (allele_index, rppcount_dict[allele_index])
                    for allele_index in possible_allele_indexes
                ),
                key=(lambda x: x[1]),
                reverse=True,
            )
            for allele_index, rppcount in allele_rppcounts_sorted[:germline_CNt]:
                allele_index_polls.setdefault(allele_index, 0)
                allele_index_polls[allele_index] += 1

        germline_allele_indexes = sorted(
            allele_index_polls.items(),
            key=(lambda x: x[1]),
            reverse=True,
        )
        germline_allele_indexes = [
            x[0] for x in germline_allele_indexes[:germline_CNt]
        ]

        # determine somatic
        possible_somatic_indexes = possible_allele_indexes.difference(germline_allele_indexes)
        somatic_allele_indexes = list()

        for sample_type, sampleid_list in sampletypes.items():
            for sampleid in sampleid_list:
                readstats = vp.readstats_dict[sampleid]
                for allele_index in possible_somatic_indexes:
                    if allele_index in somatic_allele_indexes:
                        continue
                    vaf = readstats.get_vaf(alleleclass=allele_index)
                    if vaf >= cutoff_somatic_vaf[sample_type]:
                        somatic_allele_indexes.append(allele_index)
            
        # prepare result
        result = cls(is_missing=False)
        result['germline_allele_indexes'] = germline_allele_indexes
        result['somatic_allele_indexes'] = somatic_allele_indexes

        return result

    def get_possible_allele_indexes(self):
        return set(
            itertools.chain(
                self['germline_allele_indexes'],
                self['somatic_allele_indexes'],
            )
        )

    def get_clonal_genotype(self, vp, sampleid, CNt):
        """Assumes the sample is clonal"""
        readstats = vp.readstats_dict[sampleid]

        possible_allele_indexes = self.get_possible_allele_indexes()
        rppcount_dict = {
            allele_index: readstats['rppcounts'][allele_index]
            for allele_index in possible_allele_indexes
        }
        rppcount_sum = sum(rppcount_dict.values())

        allele_vafs_sorted = sorted(
            (
                (allele_index, rppcount / rppcount_sum)
                for allele_index, rppcount in rppcount_dict.items()
            ),
            key=(lambda x: x[1]),
            reverse=True,
        )

        vaf_cutoff = (1 / CNt) * 0.4
        allele_vafs_sorted = [
            x for x in allele_vafs_sorted
            if x[1] >= vaf_cutoff
        ]

        return [x[0] for x in allele_vafs_sorted[:CNt]]
        




import textwrap
import math
import warnings

import pysam


NA_VALUES = ('', '.', None)
NA_VALUES_SET = set(NA_VALUES)


def show_info(vr):
    for key in vr.info:
        print(key, 
              get_value_info(vr, key), 
              sep='\t')


def show_format(vr):
    for sampleid in vr.header.samples:
        print('\n' + sampleid)
        for key in vr.samples[sampleid]:
            print(key, 
                  get_value_format(vr, sampleid, key), 
                  sep='\t')


def get_info(vr, key, collapse_tuple=True):
    """
    Returns:
        NA-equivalent values ("", ".", None) are converted into None
        An empty tuple is converted into None
        If the original value is a length-1 tuple, returns its element, 
            with NA conversion.
        If the original value is a tuple longer than 1, return type is tuple.
        If the original value is a tuple longer than 1 composed of only 
            NA-equivalents, returns a tuple with the same length composed 
            of None.
    """

    if key not in vr.header.info.keys():
        raise KeyError(f'Input key "{key}" is not included in the input '
                       f'variant record INFO metadata.')

    if key in vr.info:
        metadata_number = vr.header.info[key].number
        translated_number = get_translated_number_info(vr, key)
        keytype = vr.header.info[key].type
        return modify_InfoFormatValue(
            key,
            vr.info[key], 
            metadata_number,
            translated_number,
            keytype,
            for_write=False,
            typeconv=False,
            collapse_tuple=collapse_tuple,
            )
    else:
        return None

get_value_info = get_info


def get_format(vr, sampleid, key, collapse_tuple=True):
    """
    Returns:
        NA-equivalent values ("", ".", None) are converted into None
        An empty tuple is converted into None
        If the original value is a length-1 tuple, returns its element, with NA conversion.
        If the original value is a tuple longer than 1, return type is tuple.
        If the original value is a tuple longer than 1 composed of only NA-equivalents, returns a tuple with the same length composed of None.
    """

    # sanity checks
    if key == 'GT':
        raise Exception(f'"GT" is not treated with this function.')

    if key not in vr.header.formats.keys():
        raise KeyError(f'Input key "{key}" is not included in the input variant record FORMAT metadata.')

    if sampleid not in vr.header.samples:
        raise KeyError(f'Input sample ID "{sampleid}" is not included in the input variant record sample name list.')

    if key in vr.samples[sampleid]:
        metadata_number = vr.header.formats[key].number
        translated_number = get_translated_number_format(vr, key)
        keytype = vr.header.formats[key].type
        return modify_InfoFormatValue( 
            key,
            vr.samples[sampleid][key], 
            metadata_number,
            translated_number,
            keytype,
            for_write = False,
            typeconv = False,
            collapse_tuple = collapse_tuple,
            )
    else:
        return None

get_value_format = get_format


def get_genotype(vr, sampleid):
    return vr.samples[sampleid].allele_indices, vr.samples[sampleid].phased


def check_NA_info(vr, key):
    if key in vr.info:
        return check_InfoFormatValue_isNA(vr.info[key])
    else:
        return True


def check_NA_format(vr, sampleid, key):
    if sampleid not in vr.samples.keys():
        return True
    else:
        if key in vr.samples[sampleid]:
            return check_InfoFormatValue_isNA(vr.samples[sampleid][key])
        else:
            return True


def check_InfoFormatValue_isNA(val):
    """
    Args:
        val: A raw value obtained from vr.info[key] 
            or vr.samples[sampleid][key]
    """

    if isinstance(val, (list, tuple)):
        if len(val) == 0:
            """
            Value of FOO1 can be an empty tuple when:
                - Number > 1
                AND
                - INFO column looks like: FOO1;FOO2=1;FOO3=10
            """
            return True
        else:
            return set(val).issubset(NA_VALUES_SET)
    else:
        return val in NA_VALUES_SET


#############################################


def set_info(vr, key, val, typeconv=True):
    # exceptions
    #if vr.header.info[key].number == 'G':
        #warnings.warn(f'Metadata number for input key {key} is "G"; nothing is done')
    #    return

    if key not in vr.header.info:
        raise Exception(f'Key {key} is absent from INFO metadata header.')
    else:
        metadata_number = vr.header.info[key].number
        translated_number = get_translated_number_info(vr, key)
        keytype = vr.header.info[key].type
        if (metadata_number != 1) and (not isinstance(val, (list, tuple))):
            val = (val,)

        modified_val = modify_InfoFormatValue(
            key, 
            val, 
            metadata_number,
            translated_number, 
            keytype, 
            for_write=True,
            typeconv=typeconv,
            collapse_tuple=False,
            )
        vr.info[key] = modified_val

set_value_info = set_info


def set_NA_info(vr, key):
    set_value_info(vr, key, None)


def set_format(vr, sampleid, key, val, typeconv=True):
    # exceptions
    #if vr.header.formats[key].number == 'G':
        #warnings.warn(f'Metadata number for input key {key} is "G"; nothing is done')
    #    return
    if key == 'GT':
        raise Exception(f'"GT" is not treated with this function.')

    # main
    if key not in vr.header.formats:
        raise Exception(f'Key {key} is absent from FORMAT metadata header.')
    else:
        metadata_number = vr.header.formats[key].number
        translated_number = get_translated_number_format(vr, key)
        keytype = vr.header.formats[key].type
        if (metadata_number != 1) and (not isinstance(val, (list, tuple))):
            val = (val,)

        modified_val = modify_InfoFormatValue(
            key,
            val, 
            metadata_number,
            translated_number, 
            keytype, 
            for_write=True,
            typeconv=typeconv,
            collapse_tuple=False,
            )
        vr.samples[sampleid][key] = modified_val


set_value_format = set_format  # alias


def set_genotype(vr, sampleid, gt, phased):
    """
    Args:
        gt: A tuple composed of integers
        phased: True or False
    """

    vr.samples[sampleid].allele_indices = gt
    vr.samples[sampleid].phased = phased


def set_NA_format(vr, sampleid, key):
    if key == 'GT':
        set_GT_NA(vr, sampleid)
    else:
        set_value_format(vr, sampleid, key, None)


def set_GT_NA(vr, sampleid):
    vr.samples[sampleid].allele_indices = None
    vr.samples[sampleid].phased = None


def set_GT(vr, sampleid, allele_indices, phased):
    vr.samples[sampleid].allele_indices = allele_indices
    vr.samples[sampleid].phased = phased


#############################################

"""
def make_vr_for_write(vr):
    new_vr = vr.copy()
    # INFO
    for key, old_val in new_vr.info.items():
        set_value_info(new_vr, key, old_val)
    # FORMAT
    for sampleid in new_vr.header.samples:
        for key, old_val in new_vr.samples[sampleid].items():
            set_value_format(new_vr, sampleid, key, old_val)

    return new_vr
"""


def refine_vr_InfoFormatValue(vr):
    """
    Modifies each INFO or FORMAT value into
    vcf writing-compatible form. (e.g. For Type=String 
    (None,None) into ('.','.'))
    """

    # INFO
    for key, old_val in vr.info.items():
        set_value_info(vr, key, old_val)
    # FORMAT
    for sampleid in vr.header.samples:
        for key, old_val in vr.samples[sampleid].items():
            set_value_format(vr, sampleid, key, old_val)


#############################################

# not used
def get_genotype_count(vr):
    if 'GT' in vr.format.keys():
        GT_count_set = set(len(x['GT']) for x in vr.samples.values())
        if len(GT_count_set) != 1:
            raise Exception(f'Genotype counts are different between samples in this variant record:\n{vr}')
        genotype_count = GT_count_set.pop()
    else:
        genotype_count = None

    return genotype_count


def translate_metadata_number(vr, metadata_number):
    if metadata_number == 'A':
        translated_number = len(vr.alts)
    elif metadata_number == 'R':
        translated_number = len(vr.alts) + 1
    elif metadata_number == 'G':
        # combinations with repetition H(len(vr.alleles), 2)
        translated_number = math.comb(len(vr.alleles) + 2 - 1, 2)
    else:
        translated_number = metadata_number

    return translated_number


def get_translated_number_info(vr, key):
    if key in vr.header.info.keys():
        metadata_number = vr.header.info[key].number
        return translate_metadata_number(vr, metadata_number)
    else:
        return None


def get_translated_number_format(vr, key):
    if key in vr.header.formats.keys():
        metadata_number = vr.header.formats[key].number
        return translate_metadata_number(vr, metadata_number)
    else:
        return None


#############################################


def modify_InfoFormatValue(
        key, 
        val, 
        metadata_number,
        translated_number, 
        keytype, 
        for_write=False, 
        typeconv=False, 
        collapse_tuple=True,
        ):
    """
    Assumption:
        "val" is a tuple or a list if and only if metadata "Number" is 1.

    Returns:
        Type of the return value (when collapse_tuple == False):
            tuple when metadata "Number" is not 1.
            atomic when metadata "Number" is 1.

            * When collapse_tuple == True, if the return value is a 1-length
                tuple, its element is returned instead.

        Treatment of NA values:
            '', '.', and None are all treated as NA, 
                and are converted into None in the return value.

        Exceptional cases:
            If "val" is an empty tuple, it is treated as NA.
            Boolean input is returned unchanged.
            Special cases - value is unchanged for these ones:
                INFO/AS_SB_TABLE (Mutect2)
                FORMAT/GT
                Type=Flag

    Args:
        val: A raw value obtained from vr.info[key] or 
            vr.samples[sampleid][key]
        for_write: If True, modification goes for writing into a vcf file.
        keytype: VCF metadata 'Type' value. Only relevant when 
            for_write == True.
        collapse_tuple: If True, length-1 tuple is converted to its element
    """

    def handle_special_cases(key, val, keytype):
        if key == 'AS_SB_TABLE':
            # Exception for Mutect2 INFO annotation AS_SB_TABLE
            # Number == 1 but its value looks like '20,14|4,4'
            stop = True
            return_val = ','.join(val)
        elif keytype == 'Flag' or key == 'GT':
            # Flag or GT should not be modified
            stop = True
            return_val = val
        else:
            stop = False
            return_val = None

        return stop, return_val

    def get_params(val, for_write, keytype):
        if isinstance(val, (tuple, list)):
            len_val = 1 if len(val) == 0 else len(val)
        else:
            len_val = 1

        isNA = check_InfoFormatValue_isNA(val)
            # True with zero-length tuple
            # True with a non-length-zero tuple composed of NA equivalents

        if for_write:
            unifiedNA = '.' if keytype == 'String' else None
        else:
            unifiedNA = None

        return len_val, isNA, unifiedNA

    def sanity_check(key, val, metadata_number, translated_number, keytype, 
                     len_val, isNA):
        # Check if metadata Number and actual length of the value are the same
        if (
                isinstance(translated_number, int) and
                (len_val != translated_number) and
                (not isNA)):
            raise Exception(
                f'"Number" metadata and actual number of the value '
                f'does not match.\n'
                f'key = {key}, val = {val}, '
                f'translated_number = {translated_number}, '
                f'keytype = {keytype}')
        # Assumed that raw value is not a sequence type if and only if 
        # "Number" is 1.
        if (
                metadata_number == 1 and
                isinstance(val, (tuple, list))):
            raise Exception(
                f'Unexpected case: raw value is a tuple or a list'
                f'when metadata Number is 1.')
        if (
                metadata_number != 1 and
                not isinstance(val, (tuple, list))):
            raise Exception(
                f'Unexpected case: raw value is not a tuple or a list'
                f'when metadata Number is not 1.')

    def get_modified_val(
            val,
            metadata_number,
            translated_number,
            isNA,
            len_val,
            unifiedNA,
            ):
        if isNA:
            if translated_number == '.':
                modified_val = tuple([unifiedNA] * len_val)
                    # len_val is 1 when val is an empty tuple
            else:
                if metadata_number == 1:
                    modified_val = unifiedNA
                else:
                    modified_val = tuple([unifiedNA] * translated_number)
        else:
            if metadata_number == 1:
                # turn NAvalues into unifiedNA
                if val in NA_VALUES:
                    modified_val = unifiedNA
                else:
                    modified_val = val
            else:
                # turn NAvalues into unifiedNA
                modified_val = tuple(unifiedNA if x in NA_VALUES else x
                                     for x in val)

        return modified_val

    def do_typeconv(modified_val, keytype):
        if keytype == 'Flag':
            return bool(modified_val)
        else:
            if keytype == 'String' or keytype == 'Character':
                converter_func = str
            elif keytype == 'Integer':
                converter_func = int
            elif keytype == 'Float':
                converter_func = float

            def converter(val):
                return val if val in NA_VALUES else converter_func(val)

            if isinstance(modified_val, (tuple, list)):
                new_val = tuple(converter(x) for x in modified_val)
            else:
                new_val = converter(modified_val)

            return new_val

    def do_collapse_tuple(modified_val):
        if isinstance(modified_val, (tuple, list)):
            if len(modified_val) == 1:
                return modified_val[0]
            else:
                return modified_val
        else:
            return modified_val

    # main
    assert translated_number == '.' or isinstance(translated_number, int)

    stop, return_val = handle_special_cases(key, val, keytype)
    if stop:
        return return_val

    len_val, isNA, unifiedNA = get_params(val, for_write, keytype)
    sanity_check(key, val, metadata_number, translated_number, keytype, 
                 len_val, isNA)
    modified_val = get_modified_val(val, metadata_number, translated_number, 
                                    isNA, len_val, unifiedNA)
    if typeconv:
        modified_val = do_typeconv(modified_val, keytype)
    if collapse_tuple:
        modified_val = do_collapse_tuple(modified_val)

    return modified_val



import pysam

import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.variant.svcaller_parser as svcaller_parser
import handygenome.variant.varianthandler as varianthandler
import handygenome.vcfeditor.headerhandler as headerhandler
import handygenome.vcfeditor.indexing as indexing
import handygenome.variant.vcfspec as libvcfspec
import handygenome.vcfeditor.varscan_editing as varscan_editing


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser(
        description=textwrap.dedent(f"""\
            Modifies VarScan2 output file:
            - REF and ALT strings are made VCF-compatible
            - Multi-ALT lines are not split
            - non-SOMATIC, non-PASS records are removed by default
            - 
            - Original annotations are removed.

            - Split multiallelic lines.
            - Change SV caller records into breakends form.
            - Normalize non-SV records into the leftmost form.
            - Normalize SV records into the bnd1-advanced form."""
        )
    )

    # required
    parser_dict['required'].add_argument(
        '--snps',
        dest='infile_snv',
        required=True,
        help=f'The result file named "*.snps.vcf.gz"',
    )
    parser_dict['required'].add_argument(
        '--indels',
        dest='infile_indel',
        required=True,
        help=f'The result file named "*.indels.vcf.gz"',
    )

    workflow.add_outfile_arg(parser_dict['required'], required=True, must_not_exist='ask')
    workflow.add_refver_arg(parser_dict['required'], required=True, choices='all')

    # optional
    workflow.add_outfmt_arg(parser_dict['optional'], required=False)
    workflow.add_logging_args(parser_dict)

    # flag
    workflow.add_index_arg(parser_dict)
    parser_dict['flag'].add_argument(
        '--include-non-somatic',
        dest='include_non_somatic',
        action='store_true',
        help=f'If set, records in which INFO/SOMATIC flag is false are included. They are removed by default.',
    )
    parser_dict['flag'].add_argument(
        '--include-non-pass',
        dest='include_non_pass',
        action='store_true',
        help=f'If set, records in which FILTER value is not PASS are included. They are removed by default.',
    )

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    logger.info('BEGINNING')

    varscan_editing.modify_varscan_vcf(
        infile_path_list=[args.infile_snv, args.infile_indel], 
        outfile_path=args.outfile_path, 
        refver=args.refver, 
        mode_pysam=args.mode_pysam, 
        somatic_only=(not args.include_non_somatic), 
        pass_only=(not args.include_non_pass),
        index=(not args.donot_index),
    )

    logger.info('ALL SUCCESSFULLY FINISHED')



import importlib
import math
import statistics

top_package_name = __name__.split('.')[0]
module_getalleleclass_rp = importlib.import_module(top_package_name + '.ria.get_alleleclass_from_readplus')
module_getalleleclass_rpp = importlib.import_module(top_package_name + '.ria.get_alleleclass_from_readpluspair')
module_edit_readplus = importlib.import_module(top_package_name + '.ria.edit_readplus')


def add_seqREFrange_alleleclass_to_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	pos0_before_var, pos0_after_var = \
		module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	rpp.seq_REFrange = module_getalleleclass_rpp.get_seq_REFrange_from_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = pos0_before_var, 
		pos0_after_var = pos0_after_var,
		)
	rpp.alleleclass = module_getalleleclass_rpp.get_alleleclass_from_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = pos0_before_var, 
		pos0_after_var = pos0_after_var,
		)


def add_IDCinfo_to_readpluspair(rpp):
	rpp.len_ins = rpp.rp1.len_ins + rpp.rp2.len_ins
	rpp.len_del = rpp.rp1.len_del + rpp.rp2.len_del
	rpp.len_clip = rpp.rp1.len_clip + rpp.rp2.len_clip


def add_MQ(rpp):
	rpp.MQ = statistics.mean((rpp.rp1.read.mapping_quality, rpp.rp2.read.mapping_quality))


def add_MM(rpp): # mismatch
	rpp.MM = rpp.rp1.MM + rpp.rp2.MM


###

def ria_modify_readpluspair(
		rpp, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	pos0_before_var, pos0_after_var = \
		module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	# modify rpp.rp1 and rpp.rp2
	module_edit_readplus.ria_modify_readplus(
			rpp.rp1, CHROM, POS0, REF, ALT, 
			pos0_before_var = pos0_before_var, 
			pos0_after_var = pos0_after_var,
			)
	module_edit_readplus.ria_modify_readplus(
			rpp.rp2, CHROM, POS0, REF, ALT, 
			pos0_before_var = pos0_before_var, 
			pos0_after_var = pos0_after_var,
			)

	# modify rpp itself
	add_seqREFrange_alleleclass_to_readpluspair(
			rpp, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var, 
			pos0_after_var = pos0_after_var,
			)
	add_IDCinfo_to_readpluspair(rpp)
	add_MQ(rpp)
	add_MM(rpp)


def ria_modify_readpluspairlist(
		rpplist, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	pos0_before_var, pos0_after_var = \
		module_getalleleclass_rp.generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	for rpp in rpplist.rpplist:
		if not rpp.irrelevant:
			ria_modify_readpluspair(
					rpp, CHROM, POS0, REF, ALT,
				pos0_before_var = pos0_before_var, 
				pos0_after_var = pos0_after_var,
					)
	
